import argparse
import os
import sys
import traceback

from . import __version__
from .core import Client, ApplicationClient, properties, _read_driver
from .exceptions import context, SkeinError, DriverNotRunningError
from .model import ApplicationSpec, ContainerState, ApplicationState, Security
from .utils import format_table, humanize_timedelta


class _Formatter(argparse.HelpFormatter):
    """Format with a fixed argument width, due to bug in argparse measuring
    argument widths"""
    @property
    def _action_max_length(self):
        return 16

    @_action_max_length.setter
    def _action_max_length(self, value):
        pass


def fail(msg, prefix=True):
    if prefix:
        msg = 'Error: %s' % msg
    print(msg, file=sys.stderr)
    context.is_cli = False  # contextmanager skipped by SystemExit
    sys.exit(1)


def add_help(parser):
    parser.add_argument("--help", "-h", action='help',
                        help="Show this help message then exit")


def arg(*args, **kwargs):
    return (args, kwargs)


def subcommand(subparsers, name, help, *args):
    def _(func):
        parser = subparsers.add_parser(name,
                                       help=help,
                                       formatter_class=_Formatter,
                                       description=help,
                                       add_help=False)
        parser.set_defaults(func=func)
        for arg in args:
            parser.add_argument(*arg[0], **arg[1])
        add_help(parser)
        func.parser = parser
        return func
    return _


def node(subs, name, help):
    @subcommand(subs, name, help)
    def f():
        fail(f.parser.format_usage(), prefix=False)
    f.subs = f.parser.add_subparsers(metavar='command', dest='command')
    f.subs.required = True
    return f


entry = argparse.ArgumentParser(prog="skein",
                                description="Define and run YARN applications",
                                formatter_class=_Formatter,
                                add_help=False)
add_help(entry)
entry.add_argument("--version", action='version',
                   version='%(prog)s ' + __version__,
                   help="Show version then exit")
entry.set_defaults(func=lambda: fail(entry.format_usage(), prefix=False))
entry_subs = entry.add_subparsers(metavar='command', dest='command')
entry_subs.required = True

# Common arguments
app_id = arg('app_id', help='The application id', metavar='APP_ID')
app_id_or_current = arg('app_id', metavar='APP_ID',
                        help='The application id. To use in a container inside '
                             'a skein application, pass in "current"')
container_id = arg('--id', required=True,
                   help='The container id', metavar='CONTAINER_ID')


def get_driver():
    try:
        return Client.from_global_driver()
    except DriverNotRunningError:
        return Client()


def application_client_from_app_id(app_id):
    if app_id == 'current':
        return ApplicationClient.from_current()
    return get_driver().connect(app_id)


# Nodes, in order they should be in docs
application = node(entry_subs, 'application', 'Manage applications')
container = node(entry_subs, 'container', 'Manage application containers')
kv = node(entry_subs, 'kv', 'Manage the key-value store')
driver = node(entry_subs, 'driver', 'Manage the global driver')
config = node(entry_subs, 'config', 'Manage configuration')

###################
# DAEMON COMMANDS #
###################

keytab = arg('--keytab', default=None, metavar='PATH',
             help=("Path to a keytab file to use when starting the driver. "
                   "If not provided, the driver will login using the ticket "
                   "cache instead."))
principal = arg('--principal', default=None,
                help=("The principal to use when starting the driver with a "
                      "keytab."))
log = arg("--log", default=False,
          help="If provided, the driver will write logs here.")
log_level = arg("--log-level", default=None,
                help="The driver log level, default is INFO")
java_options = arg("--java-option", dest="java_options", action='append',
                   help=("Forward a Java option to the driver, can be used "
                         "multiple times."))


@subcommand(driver.subs, 'start', 'Start the skein driver',
            keytab, principal, log, log_level, java_options)
def driver_start(keytab=None, principal=None, log=False,
                 log_level=None, java_options=None):
    print(Client.start_global_driver(keytab=keytab, principal=principal,
                                     log=log, log_level=log_level,
                                     java_options=java_options))


@subcommand(driver.subs,
            'address', 'The address of the running driver')
def driver_address():
    address, _ = _read_driver()
    if address is None:
        fail("No skein driver is running")
    else:
        print(address)


@subcommand(driver.subs,
            'pid',
            'The pid of the running driver')
def driver_pid():
    _, pid = _read_driver()
    if pid is None:
        fail("No skein driver is running")
    else:
        print(pid)


@subcommand(driver.subs,
            'stop', 'Stop the skein driver',
            arg('--force', '-f', action='store_true',
                help=("Stop the process associated with the driver PID, "
                      "even if unable to verify it corresponds to a "
                      "skein driver")))
def driver_stop(force=False):
    Client.stop_global_driver(force=force)


@subcommand(driver.subs,
            'restart', 'Restart the skein driver',
            keytab, principal, log, log_level, java_options)
def driver_restart(keytab=None, principal=None, log=False, log_level=None,
                   java_options=None):
    driver_stop()
    driver_start(keytab=keytab, principal=principal,
                 log=log, log_level=log_level)


#####################
# KEYSTORE COMMANDS #
#####################

@subcommand(kv.subs,
            'get', 'Get a value from the key-value store',
            app_id_or_current,
            arg('--key',
                required=True,
                help='The key to get'),
            arg('--wait',
                action='store_true',
                help='If true, will block until the key is set'))
def kv_get(app_id, key, wait=False):
    app = application_client_from_app_id(app_id)
    result = app.kv.wait(key) if wait else app.kv[key]
    sys.stdout.buffer.write(result + b'\n')


@subcommand(kv.subs,
            'put', 'Put a value in the key-value store',
            app_id_or_current,
            arg('--key',
                required=True,
                help='The key to put'),
            arg('--value',
                required=False,
                type=lambda x: x,
                default=None,
                help='The value to put. If not provided, will be read from stdin.'))
def kv_put(app_id, key, value=None):
    if value is None:
        value = sys.stdin.buffer.read()
    else:
        value = value.encode()
    app = application_client_from_app_id(app_id)
    app.kv[key] = value


@subcommand(kv.subs,
            'del', 'Delete a value from the key-value store',
            app_id_or_current,
            arg('--key', required=True, help='The key to delete.'))
def kv_del(app_id, key):
    app = application_client_from_app_id(app_id)
    del app.kv[key]


@subcommand(kv.subs,
            'ls', 'List all keys in the key-value store',
            app_id_or_current)
def kv_ls(app_id):
    app = application_client_from_app_id(app_id)
    keys = sorted(app.kv)
    if keys:
        print('\n'.join(keys))


########################
# APPLICATION COMMANDS #
########################

def _print_application_status(apps):
    header = ['application_id', 'name', 'state', 'status', 'containers',
              'vcores', 'memory', 'runtime']
    data = [(a.id, a.name, a.state, a.final_status,
             a.usage.num_used_containers,
             a.usage.used_resources.vcores,
             a.usage.used_resources.memory,
             humanize_timedelta(a.runtime))
            for a in apps]
    print(format_table(header, data))


@subcommand(application.subs,
            'submit', 'Submit a Skein Application',
            arg('spec', help='The specification file'))
def application_submit(spec):
    if not os.path.exists(spec):
        fail("No application specification file at %r" % spec)
    try:
        spec = ApplicationSpec.from_file(spec)
    except SkeinError as exc:
        # Prettify expected errors, let rest bubble up
        fail('In file %r, %s' % (spec, exc))

    print(get_driver().submit(spec))


@subcommand(application.subs,
            'ls', 'List applications',
            arg('--all', '-a', action='store_true',
                help=('Show applications in all states (default is only '
                      'active applications)')),
            arg('--state', '-s', action='append',
                help=('Select applications with this state. May be '
                      'repeated to filter on multiple states.')),
            arg('--name', help='Select applications with this name.'),
            arg('--queue', help='Select applications with this queue.'),
            arg('--user', help='Select applications with this user.'),
            arg('--started-begin',
                help=('Select applications that started after this datetime '
                      '(inclusive). Accepts several date and time formats '
                      '(e.g. ``YYYY-M-D H:M:S`` or ``H:M``). See the '
                      'documentation for more information.')),
            arg('--started-end',
                help=('Select applications that started before this datetime '
                      '(inclusive)')),
            arg('--finished-begin',
                help=('Select applications that finished after this datetime '
                      '(inclusive)')),
            arg('--finished-end',
                help=('Select applications that finished before this datetime '
                      '(inclusive)')))
def application_ls(all=False, state=None, name=None, queue=None, user=None,
                   started_begin=None, started_end=None, finished_begin=None,
                   finished_end=None):
    if all and state is None:
        state = tuple(ApplicationState)

    apps = get_driver().get_applications(
        states=state, name=name, queue=queue, user=user,
        started_begin=started_begin, started_end=started_end,
        finished_begin=finished_begin, finished_end=finished_end
    )
    _print_application_status(apps)


@subcommand(application.subs,
            'status', 'Status of a Skein application',
            app_id)
def application_status(app_id):
    apps = get_driver().application_report(app_id)
    _print_application_status([apps])


@subcommand(application.subs,
            'logs', 'Get the logs from a completed Skein application',
            app_id,
            arg('--user', default='', type=str,
                help=('The user to get the application logs as. Requires the '
                      'current user to have permissions to proxy as ``user``. '
                      'Default is the current user.')))
def application_logs(app_id, user):
    get_driver().application_logs(app_id, user=user).dump()


@subcommand(application.subs,
            'mv', 'Move a Skein application to a different queue',
            app_id,
            arg('queue', type=str, metavar='QUEUE',
                help='The queue to move the application to.'))
def application_move(app_id, queue):
    get_driver().move_application(app_id, queue)


@subcommand(application.subs,
            'kill', 'Kill a Skein application',
            app_id,
            arg('--user', default='', type=str,
                help=('The user to kill the application as. Requires the '
                      'current user to have permissions to proxy as ``user``. '
                      'Default is the current user.')))
def application_kill(app_id, user):
    get_driver().kill_application(app_id, user=user)


@subcommand(application.subs,
            'shutdown', 'Shutdown a Skein application',
            app_id_or_current,
            arg('--status', default='SUCCEEDED',
                help='Final Application Status. Default is SUCCEEDED'),
            arg('--diagnostics', default=None,
                help=('The application diagnostic exit message. If not '
                      'provided, a default will be used.')))
def application_shutdown(app_id, status, diagnostics):
    application_client_from_app_id(app_id).shutdown(status, diagnostics)


@subcommand(application.subs,
            'specification', 'Get specification for a running skein application',
            app_id_or_current)
def application_specification(app_id):
    app = application_client_from_app_id(app_id)
    print(app.get_specification()
             .to_yaml(skip_nulls=True))


######################
# CONTAINER COMMANDS #
######################

def _print_container_status(containers):
    header = ['service', 'id', 'state', 'runtime']
    data = [(c.service_name, c.id, c.state, humanize_timedelta(c.runtime))
            for c in containers]
    print(format_table(header, data))


@subcommand(container.subs,
            'ls', 'List containers',
            app_id_or_current,
            arg('--all', '-a', action='store_true',
                help='Show all containers (default is only active containers)'),
            arg("--service", action='append',
                help=('Filter by container services. May be repeated '
                      'to select multiple services.')),
            arg("--state", action='append',
                help=('Filter by container states. May be repeated '
                      'to select multiple states.')))
def container_ls(app_id, all=False, service=None, state=None):
    app = application_client_from_app_id(app_id)
    if all and state is None:
        state = tuple(ContainerState)
    containers = app.get_containers(states=state, services=service)
    _print_container_status(containers)


@subcommand(container.subs,
            'kill', 'Kill a container',
            app_id_or_current,
            container_id)
def container_kill(app_id, id):
    application_client_from_app_id(app_id).kill_container(id)


@subcommand(container.subs,
            'scale', 'Scale a service to a requested number of containers',
            app_id_or_current,
            arg('--service', '-s', required=True, help='Service name'),
            arg('--number', '-n', type=int, required=True,
                help='The requested number of instances'))
def container_scale(app_id, service, number):
    application_client_from_app_id(app_id).scale(service, number)


##################
# CONFIG COMMAND #
##################

@subcommand(config.subs,
            'gencerts',
            'Generate security credentials. Creates a self-signed TLS '
            'key/certificate pair for securing Skein communication, and writes '
            'it to the skein configuration directory ("~.skein/" by default).',
            arg('--force', '-f', action='store_true',
                help='Overwrite existing configuration'))
def config_gencerts(force=False):
    sec = Security.new_credentials()
    sec.to_directory(directory=properties.config_dir, force=force)


def main(args=None):
    kwargs = vars(entry.parse_args(args=args))
    kwargs.pop('command', None)  # Drop unnecessary `command` arg
    func = kwargs.pop('func')
    try:
        with context.set_cli():
            func(**kwargs)
    except KeyError as exc:
        fail("Key %s is not set" % str(exc))
    except SkeinError as exc:
        fail(str(exc))
    except Exception:
        fail("Unexpected Error:\n%s" % traceback.format_exc(), prefix=False)
    sys.exit(0)


if __name__ == '__main__':
    main()



import warnings
import sys
from contextlib import contextmanager

__all__ = ('FileExistsError',    # py2 compat
           'FileNotFoundError',  # py2 compat
           'SkeinError',
           'ConnectionError',
           'TimeoutError',
           'DriverNotRunningError',
           'ApplicationNotRunningError',
           'DriverError',
           'ApplicationError')


FileExistsError = FileExistsError
FileNotFoundError = FileNotFoundError


class SkeinError(Exception):
    """Base class for Skein specific exceptions"""


class ConnectionError(SkeinError, ConnectionError):
    """Failed to connect to the driver or application master"""


class TimeoutError(SkeinError, TimeoutError):
    """Request to driver or application master timed out"""


class DriverNotRunningError(ConnectionError):
    """The driver process is not currently running"""


class ApplicationNotRunningError(ConnectionError):
    """The application master is not currently running"""


class DriverError(SkeinError):
    """Internal exceptions from the driver"""


class ApplicationError(SkeinError):
    """Internal exceptions from the application master"""


class _Context(object):
    def __init__(self):
        self.is_cli = False

    def warn(self, msg):
        if self.is_cli:
            print(msg + "\n", file=sys.stderr)
        else:
            warnings.warn(msg)

    @contextmanager
    def set_cli(self):
        old = self.is_cli
        self.is_cli = True
        yield
        self.is_cli = old

    @classmethod
    def register_wrapper(cls, typ):
        name = typ.__name__
        typ2 = type(name, (typ, SkeinError), {})

        def wrap(self, msg):
            return typ2(msg) if self.is_cli else typ(msg)

        setattr(cls, name, wrap)


for exc in [ValueError, KeyError, TypeError, FileExistsError,
            FileNotFoundError]:
    _Context.register_wrapper(exc)


context = _Context()



import datetime
import os
import pickle
import subprocess
import time
import warnings
import weakref
from contextlib import closing
from multiprocessing.pool import Pool

import pytest

import skein
from skein.core import Properties
from skein.exceptions import FileNotFoundError, FileExistsError
from skein.utils import pid_exists
from skein.test.conftest import (run_application, wait_for_containers,
                                 wait_for_completion, get_logs, KEYTAB_PATH,
                                 HADOOP3)


def test_properties():
    assert len(skein.properties) == len(dict(skein.properties))
    assert skein.properties.config_dir == skein.properties['config_dir']
    assert 'config_dir' in dir(skein.properties)
    assert 'missing' not in skein.properties

    with pytest.raises(AttributeError):
        skein.properties.missing

    with pytest.raises(AttributeError):
        skein.properties.missing = 1

    with pytest.raises(KeyError):
        skein.properties['missing']

    with pytest.raises(TypeError):
        skein.properties['missing'] = 1


def test_security(tmpdir):
    path = str(tmpdir)
    s1 = skein.Security.new_credentials()
    s2 = s1.to_directory(path)
    s3 = skein.Security.from_directory(path)
    assert s1 != s2
    assert s2 == s3

    with pytest.raises(FileExistsError):
        s1.to_directory(path)

    s3 = skein.Security.new_credentials()
    s3.to_directory(path, force=True)

    cert_path = os.path.join(path, 'skein.crt')

    with open(cert_path, 'rb') as fil:
        cert3 = fil.read()
    assert s3.cert_bytes == cert3
    assert s3.cert_bytes != s1.cert_bytes

    os.remove(cert_path)
    with pytest.raises(FileNotFoundError):
        skein.Security.from_directory(path)


def test_security_get_bytes(tmpdir):
    path = str(tmpdir)
    s1 = skein.Security.new_credentials()
    s2 = s1.to_directory(path)

    for kind in ['cert', 'key']:
        x1 = s1._get_bytes(kind)
        x2 = s2._get_bytes(kind)
        assert isinstance(x1, bytes)
        assert isinstance(x2, bytes)
        assert x1 == x2

    not_local = skein.Security(cert_file="hdfs:///some/path",
                               key_file="hdfs:///some/path")
    missing = skein.Security(cert_file="definitely/a/missing/path",
                             key_file="definitely/a/missing/path")

    for kind in ['cert', 'key']:
        with pytest.raises(ValueError):
            not_local._get_bytes(kind)

        with pytest.raises(FileNotFoundError):
            missing._get_bytes(kind)


def test_security_auto_inits(skein_config):
    with pytest.warns(None) as rec:
        sec = skein.Security.from_default()

    assert len(rec) == 1
    assert 'Skein global security credentials not found' in str(rec[0])

    cert_path = os.path.join(skein.properties.config_dir, 'skein.crt')
    key_path = os.path.join(skein.properties.config_dir, 'skein.crt')

    assert os.path.exists(cert_path)
    assert os.path.exists(key_path)

    with pytest.warns(None) as rec:
        sec2 = skein.Security.from_default()

    assert not rec
    assert sec == sec2


def _pool_initializer(config_dir):
    skein.properties._mapping['config_dir'] = config_dir


def _from_default_count_warnings(n):
    with warnings.catch_warnings(record=True) as rec:
        skein.Security.from_default()
    return len(rec)


def test_security_auto_inits_no_race_condition(tmpdir):
    config_dir = str(tmpdir)
    pool = Pool(processes=4, initializer=_pool_initializer,
                initargs=(config_dir,))
    with closing(pool):
        num_warnings = sum(pool.map(_from_default_count_warnings, range(8)))
    # Only one process actually writes (and raises a warning)
    # all others just read in credentials
    assert num_warnings == 1


def test_client(security, kinit, tmpdir):
    logpath = str(tmpdir.join("log.txt"))

    with skein.Client(security=security, log=logpath) as client:
        # smoketests
        client.get_applications()
        repr(client)

        client2 = skein.Client(address=client.address, security=security)
        assert client2._proc is None

        # smoketests
        client2.get_applications()
        repr(client2)

    # Process was definitely closed
    assert not pid_exists(client._proc.pid)

    # no-op to call close again
    client.close()

    # Log was written
    assert os.path.exists(logpath)
    with open(logpath) as fil:
        assert len(fil.read()) > 0

    # Connection error on closed client
    with pytest.raises(skein.ConnectionError):
        client2.get_applications()

    # Connection error on connecting to missing driver
    with pytest.raises(skein.ConnectionError):
        skein.Client(address=client.address, security=security)


def test_client_closed_when_reference_dropped(security, kinit):
    client = skein.Client(security=security, log=False)
    ref = weakref.ref(client)

    pid = client._proc.pid

    del client
    assert ref() is None
    assert not pid_exists(pid)


def test_client_errors_nicely_if_not_logged_in(security, not_logged_in):
    appid = 'application_1526134340424_0012'

    spec = skein.ApplicationSpec(
        name="should_never_get_to_run",
        queue="default",
        services={
            'service': skein.Service(
                resources=skein.Resources(memory=32, vcores=1),
                script='env')
        }
    )

    with skein.Client(security=security) as client:
        for func, args in [('get_applications', ()),
                           ('get_nodes', ()),
                           ('get_queue', ('default',)),
                           ('get_child_queues', ('default',)),
                           ('get_all_queues', ()),
                           ('application_report', (appid,)),
                           ('application_logs', (appid,)),
                           ('connect', (appid,)),
                           ('move_application', (appid, 'default')),
                           ('kill_application', (appid,)),
                           ('submit', (spec,))]:
            with pytest.raises(skein.DriverError) as exc:
                getattr(client, func)(*args)
            assert 'kinit' in str(exc.value)


def test_client_starts_without_java_home(monkeypatch, tmpdir, security, kinit):
    monkeypatch.delenv('JAVA_HOME', raising=False)

    logpath = str(tmpdir.join("log.txt"))

    with skein.Client(security=security, log=logpath) as client:
        # do an operation to ensure everything is working
        client.get_applications()

    with open(logpath) as fil:
        data = fil.read()
        assert 'WARN' not in data
        assert 'native-hadoop' not in data


def test_client_set_log_level(security, kinit, tmpdir):
    logpath = str(tmpdir.join("log.txt"))

    with skein.Client(security=security, log=logpath, log_level='debug') as client:
        # do an operation to ensure everything is working
        client.get_applications()

    with open(logpath) as fil:
        data = fil.read()
        assert 'DEBUG' in data


@pytest.mark.parametrize('use_env', [True, False])
def test_client_forward_java_options(use_env, security, kinit, tmpdir, monkeypatch):
    logpath = str(tmpdir.join("log.txt"))

    if use_env:
        monkeypatch.setenv('SKEIN_DRIVER_JAVA_OPTIONS',
                           '-Dskein.log.level=debug')
        kwargs = {}
    else:
        kwargs = {'java_options': ['-Dskein.log.level=debug']}

    with skein.Client(security=security, log=logpath, **kwargs) as client:
        # do an operation to ensure everything is working
        client.get_applications()

    with open(logpath) as fil:
        data = fil.read()
        assert 'DEBUG' in data


def test_client_login_from_keytab(security, not_logged_in):
    with skein.Client(principal='testuser', keytab=KEYTAB_PATH,
                      security=security) as client:
        # login worked
        client.get_applications()

    # Improper principal/keytab pair
    with pytest.raises(skein.DriverError):
        skein.Client(principal='not_the_right_user', keytab=KEYTAB_PATH,
                     security=security)

    # Keytab file missing
    with pytest.raises(FileNotFoundError):
        skein.Client(principal='testuser', keytab='/not/a/real/path',
                     security=security)

    # Must specify both principal and keytab
    with pytest.raises(ValueError):
        skein.Client(principal='testuser', security=security)

    with pytest.raises(ValueError):
        skein.Client(keytab=KEYTAB_PATH, security=security)


def test_client_picklable(client):
    b = pickle.dumps(client)
    client2 = pickle.loads(b)
    assert client2.address == client.address
    assert client2.security == client.security
    assert client2._proc is None
    # Smoketest
    client2.get_applications()


def test_get_nodes(client):
    nodes = client.get_nodes()
    assert nodes

    # Doesn't exist for Hadoop <= 2.8, should be empty list
    nodes = client.get_nodes(states=['SHUTDOWN'])
    assert not nodes

    # Should still have results here
    nodes = client.get_nodes(states=['SHUTDOWN', 'RUNNING'])
    assert nodes


def test_get_queue_methods(client):
    def qnames(qs):
        return {q.name for q in qs}

    all_queues = client.get_all_queues()
    assert qnames(all_queues) == {'default', 'fruit', 'apples', 'bananas', 'oranges'}

    assert qnames(client.get_child_queues('root')) == {'default', 'fruit'}
    assert client.get_child_queues('default') == []

    assert client.get_queue('default').name == 'default'

    with pytest.raises(ValueError):
        client.get_queue("missing")

    with pytest.raises(ValueError):
        client.get_child_queues("missing")


@pytest.fixture
def at_least_3_apps_in_history(client):
    all_apps = client.get_applications(states=skein.model.ApplicationState.values())
    all_expected = [a for a in all_apps if a.user == 'testuser' and a.queue == 'default']
    for _ in range(3 - len(all_expected)):
        with run_application(client) as app:
            app.shutdown()


def test_get_applications(client, at_least_3_apps_in_history):
    all_states = skein.model.ApplicationState.values()
    all_apps = client.get_applications(states=all_states)
    # This test is a bit tricky, since the many apps may have been run
    # previously, some not in our test suite. Here we filter to just ones we'd
    # expect.
    all_expected = [a for a in all_apps if a.user == 'testuser' and a.queue == 'default']
    assert all_expected

    def assert_eq(a, b):
        a = set(i.id for i in a)
        b = set(i.id for i in b)
        assert a == b

    # Filter on user and queue to show identical
    common = {'states': all_states, 'user': 'testuser', 'queue': 'default'}
    res = client.get_applications(**common)
    assert_eq(res, all_expected)

    # Filter on name as well
    app_name = all_expected[0].name
    sol = [a for a in all_expected if a.name == app_name]
    res = client.get_applications(name=app_name, **common)
    assert_eq(res, sol)

    # Filter out all apps
    assert not client.get_applications(states=all_states, user='not-a-real-value')
    assert not client.get_applications(states=all_states, queue='not-a-real-value')
    assert not client.get_applications(states=all_states, name='not-a-real-value')

    mid_app = all_expected[len(all_expected) // 2]
    start_time = mid_app.start_time.replace(microsecond=0)
    finish_time = mid_app.finish_time.replace(microsecond=0)

    # All apps started after start_time
    sol = [a for a in all_expected if a.start_time >= start_time]
    res = client.get_applications(started_begin=start_time, **common)
    assert_eq(res, sol)

    # All apps started before start_time
    sol = [a for a in all_expected if a.start_time <= start_time]
    res = client.get_applications(started_end=start_time, **common)
    assert_eq(res, sol)

    # All apps finished after finish_time
    sol = [a for a in all_expected if a.finish_time >= finish_time]
    res = client.get_applications(finished_begin=finish_time, **common)
    assert_eq(res, sol)

    # All apps finished before finish_time
    sol = [a for a in all_expected if a.finish_time <= finish_time]
    res = client.get_applications(finished_end=finish_time, **common)
    assert_eq(res, sol)

    # Check time parsing
    t = datetime.datetime.now()
    for format in ['%Y-%m-%d', '%Y-%m-%d %H:%M', '%Y-%m-%d %H:%M:%S']:
        t_str = t.strftime(format)
        sol = datetime.datetime.strptime(t_str, format)
        res = client._parse_datetime(t_str, '')
        assert sol == res

    for format in ['%H:%M', '%H:%M:%S']:
        d = datetime.datetime(year=t.year, month=t.month, day=t.day)
        t_str = t.strftime(format)
        tm = datetime.datetime.strptime(t_str, format).time()
        sol = datetime.datetime.combine(d, tm)
        res = client._parse_datetime(t_str, '')
        assert sol == res

    sol = [a for a in all_expected if a.finish_time <= finish_time]
    res = client.get_applications(
        finished_end=finish_time.strftime('%Y-%m-%d %H:%M:%S'), **common
    )
    assert_eq(res, sol)

    with pytest.raises(TypeError) as exc:
        res = client.get_applications(finished_end=123)
    assert 'finished_end' in str(exc.value)

    with pytest.raises(ValueError) as exc:
        res = client.get_applications(finished_end='4-1-2019')
    assert 'finished_end' in str(exc.value)


def test_appclient_and_security_in_container(monkeypatch, tmpdir, security):
    # Not running in a container
    with pytest.raises(ValueError) as exc:
        skein.ApplicationClient.from_current()
    assert str(exc.value) == "Not running inside a container"

    # Patch environment variables so it looks like a container
    app_id = 'application_1526134340424_0012'
    container_id = 'container_1526134340424_0012_01_000005'
    address = 'edge.example.com:8765'
    bad_dir = tmpdir.mkdir("nothing_in_here")
    bad_dir.mkdir(container_id)

    for key, val in [('SKEIN_APPLICATION_ID', app_id),
                     ('CONTAINER_ID', container_id),
                     ('SKEIN_APPMASTER_ADDRESS', address),
                     ('LOCAL_DIRS', str(bad_dir))]:
        monkeypatch.setenv(key, val)

    properties = Properties()
    assert properties.container_dir is None
    monkeypatch.setattr(skein.core, 'properties', properties)

    # In container, but unable to find security configuration
    with pytest.raises(FileNotFoundError) as exc:
        skein.ApplicationClient.from_current()
    assert str(exc.value) == "Failed to resolve .skein.{crt,pem} in 'LOCAL_DIRS'"

    with pytest.raises(FileNotFoundError) as exc:
        skein.Security.from_default()
    assert str(exc.value) == "Failed to resolve .skein.{crt,pem} in 'LOCAL_DIRS'"

    # Add proper LOCAL_DIRS environment
    good_dir = tmpdir.mkdir('good_dir')
    local_dir = good_dir.mkdir(container_id)
    with open(str(local_dir.join(".skein.crt")), 'wb') as fil:
        fil.write(security._get_bytes('cert'))
    with open(str(local_dir.join(".skein.pem")), 'wb') as fil:
        fil.write(security._get_bytes('key'))
    monkeypatch.setenv('LOCAL_DIRS', '%s,%s' % (bad_dir, good_dir))

    properties = Properties()
    assert properties.container_dir == local_dir
    monkeypatch.setattr(skein.core, 'properties', properties)

    # Picks up full configuration from environment variables
    app = skein.ApplicationClient.from_current()
    assert app.id == app_id
    assert app.address == address

    security2 = skein.Security.from_default()
    assert security._get_bytes('key') == security2._get_bytes('key')
    assert security._get_bytes('cert') == security2._get_bytes('cert')


def test_simple_app(client):
    with run_application(client) as app:
        # smoketest repr
        repr(app)

        # Test get_specification
        a = app.get_specification()
        assert isinstance(a, skein.ApplicationSpec)
        assert 'sleeper' in a.services

        assert client.application_report(app.id).state == 'RUNNING'

        # Connect with a new client
        with client.connect(app.id) as app2:
            app2.get_specification()

        # application client is closed
        with pytest.raises(Exception):
            app2.get_specification()

        # Closing an application client is idempotent
        app2.close()

        app.shutdown()

    with pytest.raises(skein.ConnectionError):
        client.connect(app.id)

    with pytest.raises(skein.ConnectionError):
        client.connect(app.id, wait=False)

    # On Travis CI there can be some lag between application being shutdown and
    # application actually shutting down. Retry up to 5 seconds before failing.
    with pytest.raises(skein.ConnectionError):
        timeout = 5
        while timeout:
            try:
                app.get_specification()
            except skein.ConnectionError:
                raise
            else:
                # Didn't fail, try again later
                time.sleep(0.1)
                timeout -= 0.1

    running_apps = client.get_applications()
    assert app.id not in {a.id for a in running_apps}

    finished_apps = client.get_applications(states=['finished'])
    assert app.id in {a.id for a in finished_apps}


def test_shutdown_arguments(client):
    status = 'killed'
    diagnostics = 'This is a test diagnostic message'

    with run_application(client) as app:
        app.shutdown(status, diagnostics)
        assert wait_for_completion(client, app.id) == 'KILLED'

    # There's a noticeable lag in the YARN resource manager between an
    # application being marked as finished and its diagnostics message being
    # updated. Retry up to 5 seconds before failing.
    timeout = 5
    while timeout:
        report = client.application_report(app.id)
        if report.diagnostics:
            break
        time.sleep(0.1)
        timeout -= 0.1
    assert report.diagnostics == diagnostics
    assert report.final_status == status


def test_dynamic_containers(client):
    spec = skein.ApplicationSpec(
        name="test_dynamic_containers",
        services={
            'sleeper': skein.Service(
                instances=1,
                resources=skein.Resources(memory=32, vcores=1),
                script='sleep infinity'
            )
        },
        master=skein.Master(script='sleep infinity')
    )
    with run_application(client, spec=spec) as app:
        initial = wait_for_containers(app, 1, states=['RUNNING'])
        assert initial[0].state == 'RUNNING'
        assert initial[0].service_name == 'sleeper'

        # Scale sleepers up to 3 containers
        new = app.scale('sleeper', 3)
        assert len(new) == 2
        for c in new:
            assert c.state == 'REQUESTED'
        wait_for_containers(app, 3, services=['sleeper'], states=['RUNNING'])

        # Scale down to 1 container
        stopped = app.scale('sleeper', 1)
        assert len(stopped) == 2
        # Stopped oldest 2 instances
        assert stopped[0].instance == 0
        assert stopped[1].instance == 1

        # Scale up to 2 containers
        new = app.scale('sleeper', 2)
        # Calling twice is no-op
        new2 = app.scale('sleeper', 2)
        assert len(new2) == 0
        assert new[0].instance == 3
        current = wait_for_containers(app, 2, services=['sleeper'],
                                      states=['RUNNING'])
        assert current[0].instance == 2
        assert current[1].instance == 3

        # Manually kill instance 3
        app.kill_container('sleeper_3')
        current = app.get_containers()
        assert len(current) == 1
        assert current[0].instance == 2

        # Fine to kill already killed container
        app.kill_container('sleeper_1')

        # All killed containers
        killed = app.get_containers(states=['killed'])
        assert len(killed) == 3
        assert [c.instance for c in killed] == [0, 1, 3]
        # All completed containers have an exit message
        assert all(c.exit_message for c in killed)

        # Add containers by delta
        ncurrent = len(app.get_containers())
        new = app.scale('sleeper', delta=2)
        assert len(new) == 2
        assert len(app.get_containers()) == ncurrent + 2

        # Remove containers by delta
        ncurrent = len(app.get_containers())
        assert ncurrent >= 1
        res = app.scale('sleeper', delta=-1)
        assert len(res) == 1
        assert len(app.get_containers()) == ncurrent - 1

        # Removing more containers than active removes all containers
        ncurrent = len(app.get_containers())
        res = app.scale('sleeper', delta=-(ncurrent + 2))
        assert len(res) == ncurrent
        assert len(app.get_containers()) == 0

        # Can't specify both count and delta
        with pytest.raises(ValueError):
            app.scale('sleeper', count=2, delta=2)

        # Must specify either count or delta
        with pytest.raises(ValueError):
            app.scale('sleeper')

        # Can't scale non-existant service
        with pytest.raises(ValueError):
            app.scale('foobar', 2)

        # Can't scale negative
        with pytest.raises(ValueError):
            app.scale('sleeper', -5)

        # Can't kill non-existant container
        with pytest.raises(ValueError):
            app.kill_container('foobar_1')

        with pytest.raises(ValueError):
            app.kill_container('sleeper_500')

        # Invalid container id
        with pytest.raises(ValueError):
            app.kill_container('fooooooo')

        # Can't get containers for non-existant service
        with pytest.raises(ValueError):
            app.get_containers(services=['sleeper', 'missing'])

        app.shutdown()


def test_add_container(client):
    script = ('echo "$SKEIN_CONTAINER_ID - MYENV=$MYENV"\n'
              'echo "$SKEIN_CONTAINER_ID - MYENV2=$MYENV2"\n'
              'if [[ "$MYENV" == "bar" ]]; then\n'
              '  exit 1\n'
              'else\n'
              '  exit 0\n'
              'fi')

    spec = skein.ApplicationSpec(
        name="test_add_container",
        master=skein.Master(script="sleep infinity"),
        services={
            'test': skein.Service(
                instances=0,
                resources=skein.Resources(memory=32, vcores=1),
                env={'MYENV': 'foo',
                     'MYENV2': 'baz'},
                max_restarts=1,
                script=script
            )
        }
    )

    with run_application(client, spec=spec) as app:
        # Add container with new overrides
        c = app.add_container('test')
        assert c.instance == 0
        wait_for_containers(app, 1, states=['RUNNING', 'SUCCEEDED'])

        # Non-existant service
        with pytest.raises(ValueError):
            app.add_container('foobar')

        # Add container with override for MYENV
        c = app.add_container('test', {'MYENV': 'bar'})
        assert c.instance == 1

        # The new env var triggers a failure, should fail twice,
        # then fail the whole application
        assert wait_for_completion(client, app.id) == 'FAILED'

    logs = get_logs(client, app.id)
    assert "test_0 - MYENV=foo" in logs
    assert "test_0 - MYENV2=baz" in logs

    assert "test_1 - MYENV=bar" in logs
    assert "test_1 - MYENV2=baz" in logs

    assert "test_2 - MYENV=bar" in logs
    assert "test_2 - MYENV2=baz" in logs

    assert "test_3" not in logs


@pytest.mark.parametrize('runon', ['service', 'master'])
def test_container_environment(runon, client, has_kerberos_enabled):
    script = ('set -e\n'
              'env\n'
              'echo "LOGIN_ID=[$(whoami)]"\n'
              'hdfs dfs -touchz /user/testuser/test_container_permissions\n'
              'yarn application -list')
    kwargs = dict(resources=skein.Resources(memory=512, vcores=1),
                  script=script)
    services = master = None
    if runon == 'service':
        services = {'service': skein.Service(**kwargs)}
    else:
        master = skein.Master(**kwargs)

    spec = skein.ApplicationSpec(name="test_container_permissions_%s" % runon,
                                 queue="default",
                                 services=services,
                                 master=master)

    with run_application(client, spec=spec, connect=False) as app_id:
        assert wait_for_completion(client, app_id) == 'SUCCEEDED'

    logs = get_logs(client, app_id)
    assert "USER=testuser" in logs
    assert 'SKEIN_APPMASTER_ADDRESS=' in logs
    assert 'SKEIN_APPLICATION_ID=%s' % app_id in logs
    if runon == 'service':
        assert 'SKEIN_CONTAINER_ID=service_0' in logs
    assert 'SKEIN_RESOURCE_MEMORY=512' in logs
    assert 'SKEIN_RESOURCE_VCORES=1' in logs
    assert 'CLASSPATH' not in logs

    if has_kerberos_enabled:
        assert "LOGIN_ID=[testuser]" in logs
        assert "HADOOP_USER_NAME" not in logs
    else:
        assert "LOGIN_ID=[yarn]" in logs
        assert "HADOOP_USER_NAME" in logs


def test_file_systems(client):
    script = 'hdfs dfs -touchz /user/testuser/test_file_systems'
    service = skein.Service(resources=skein.Resources(memory=128, vcores=1),
                            script=script)
    spec = skein.ApplicationSpec(name="test_file_systems",
                                 queue="default",
                                 services={'service': service},
                                 file_systems=["hdfs://master.example.com:9000"])

    with run_application(client, spec=spec) as app:
        assert wait_for_completion(client, app.id) == 'SUCCEEDED'


@pytest.mark.parametrize('use_skein', [True, False])
def test_kill_application_removes_appdir(use_skein, client):
    hdfs = pytest.importorskip('pyarrow.hdfs')

    with run_application(client) as app:
        if use_skein:
            client.kill_application(app.id)
        else:
            subprocess.check_call(["yarn", "application", "-kill", app.id])

    fs = hdfs.connect()
    assert not fs.exists("/user/testuser/.skein/%s" % app.id)


def test_submit_failure_removes_appdir(client):
    hdfs = pytest.importorskip('pyarrow.hdfs')

    # Application with vcores > max
    spec = skein.ApplicationSpec(
        name="test_submit_failure_removes_appdir",
        queue="default",
        master=skein.Master(
            resources=skein.Resources(vcores=1000, memory=32),
            script="echo 'should never run'"
        )
    )

    fs = hdfs.connect()
    before = set(fs.ls("/user/testuser/.skein"))

    with pytest.raises(skein.DriverError):
        client.submit(spec)

    after = set(fs.ls("/user/testuser/.skein"))

    # Application directory is cleaned up on failure
    assert before == after


custom_log4j_properties = """
# Root logger option
log4j.rootCategory=INFO, console

# Redirect log messages to console
log4j.appender.console=org.apache.log4j.ConsoleAppender
log4j.appender.console.Target=System.out
log4j.appender.console.layout=org.apache.log4j.PatternLayout
log4j.appender.console.layout.ConversionPattern=CUSTOM-LOG4J-SUCCEEDED %m
"""


def test_custom_log4j_properties(client, tmpdir):
    configpath = str(tmpdir.join("log4j.properties"))
    service = skein.Service(resources=skein.Resources(memory=32, vcores=1),
                            script='ls')
    spec = skein.ApplicationSpec(name="test_custom_log4j_properties",
                                 queue="default",
                                 master=skein.Master(log_config=configpath),
                                 services={'service': service})
    with open(configpath, 'w') as f:
        f.write(custom_log4j_properties)

    with run_application(client, spec=spec) as app:
        assert wait_for_completion(client, app.id) == 'SUCCEEDED'

    logs = get_logs(client, app.id)
    assert 'CUSTOM-LOG4J-SUCCEEDED' in logs


def test_set_log_level(client):
    service = skein.Service(resources=skein.Resources(memory=32, vcores=1),
                            script='ls')
    spec = skein.ApplicationSpec(name="test_custom_log4j_properties",
                                 queue="default",
                                 master=skein.Master(log_level='debug'),
                                 services={'service': service})

    with run_application(client, spec=spec) as app:
        assert wait_for_completion(client, app.id) == 'SUCCEEDED'

    logs = get_logs(client, app.id)
    assert 'DEBUG' in logs


@pytest.mark.parametrize('kind', [
    pytest.param('master', marks=pytest.mark.xfail(
        reason="Memory error on master isn't deterministic"
    )),
    'service'
])
def test_memory_limit_exceeded(kind, client):
    resources = skein.Resources(memory=32, vcores=1)
    # Allocate noticeably more memory than the 32 MB limit
    script = 'python -c "b = bytearray(int(256e6)); import time; time.sleep(10)"'

    master = services = None
    if kind == 'master':
        master = skein.Master(resources=resources, script=script)
        search_txt = "memory limit"
    else:
        services = {
            'service': skein.Service(resources=resources, script=script)
        }
        search_txt = "memory used"
    spec = skein.ApplicationSpec(name="test_memory_limit_exceeded_%s" % kind,
                                 queue="default",
                                 master=master,
                                 services=services)
    with run_application(client, spec=spec, connect=False) as app_id:
        assert wait_for_completion(client, app_id) == "FAILED"
    logs = get_logs(client, app_id)
    assert search_txt in logs

    if kind == 'master':
        report = client.application_report(app_id)
        assert 'memory limit' in report.diagnostics


@pytest.mark.parametrize('strict', [False, True])
def test_node_locality(client, strict):
    if strict:
        relax_locality = False
        nodes = ['worker.example.com']
        racks = []
    else:
        relax_locality = True
        nodes = ['not.a.real.host.name']
        racks = ['not.a.real.rack.name']

    service = skein.Service(
        resources=skein.Resources(memory=32, vcores=1),
        script='sleep infinity',
        nodes=nodes,
        racks=racks,
        relax_locality=relax_locality
    )
    spec = skein.ApplicationSpec(name="test_node_locality",
                                 queue="default",
                                 services={"service": service})
    with run_application(client, spec=spec) as app:
        wait_for_containers(app, 1, states=['RUNNING'])
        spec2 = app.get_specification()
        app.shutdown()

    service2 = spec2.services['service']
    assert service2.nodes == nodes
    assert service2.racks == racks
    assert service2.relax_locality == relax_locality


def test_set_application_progress(client):
    with run_application(client) as app:
        app.set_progress(0.5)
        # The report won't update until the allocator loop runs. Try a couple
        # times.
        for _ in range(6):
            report = client.application_report(app.id)
            if report.progress == 0.5:
                break
            time.sleep(1)
        else:
            assert report.progress == 0.5

        with pytest.raises(ValueError):
            app.set_progress(-0.5)

        with pytest.raises(ValueError):
            app.set_progress(1.5)

        app.shutdown()


def test_proxy_user(client):
    hdfs = pytest.importorskip('pyarrow.hdfs')

    spec = skein.ApplicationSpec(
        name="test_proxy_user",
        user="alice",
        services={
            "service": skein.Service(
                resources=skein.Resources(memory=32, vcores=1),
                script="sleep infinity")
        }
    )
    with run_application(client, spec=spec) as app:
        spec2 = app.get_specification()
        client.kill_application(app.id, user="alice")

    # Alice used throughout process
    assert spec2.user == 'alice'
    for fil in spec2.services['service'].files.values():
        assert fil.source.startswith('hdfs://master.example.com:9000/user/alice')

    # Can get logs as user
    logs = get_logs(client, app.id, user="alice")
    assert app.id in logs
    assert "application.master.log" in logs

    # Application directory deleted after kill
    fs = hdfs.connect()
    assert not fs.exists("/user/testuser/.skein/%s" % app.id)


def test_proxy_user_no_permissions(client, hadoop3):
    if hadoop3:
        pytest.skip("Lack of proxyuser permissions causes "
                    "yarnclient to hang in hadoop3")
    spec = skein.ApplicationSpec(
        name="test_proxy_user_no_permissions",
        user="bob",
        services={
            'service': skein.Service(
                resources=skein.Resources(memory=32, vcores=1),
                script='env')
        }
    )
    # No permission to submit as user
    with pytest.raises(skein.DriverError) as exc:
        client.submit(spec)

    exc_msg = str(exc.value)
    assert 'testuser' in exc_msg
    assert 'bob' in exc_msg


def test_security_specified(client):
    security = skein.Security.new_credentials()
    spec = skein.ApplicationSpec(
        name="test_security_specified",
        master=skein.Master(security=security,
                            script='sleep infinity')
    )
    with run_application(client, spec=spec) as app:
        assert app.security is security
        assert app.security != client.security

        spec2 = app.get_specification()

        app2 = client.connect(app.id, security=security)
        # Smoketest, can communicate
        app2.get_specification()

        app3 = client.connect(app.id)
        with pytest.raises(skein.ConnectionError):
            # Improper security credentials
            app3.get_specification()

        app.shutdown()

    remote_security = spec2.master.security
    assert remote_security.cert_bytes is None
    assert remote_security.key_bytes is None
    assert remote_security.cert_file.source.startswith('hdfs')
    assert remote_security.key_file.source.startswith('hdfs')


def test_master_driver_foo(client, tmpdir):
    filpath = str(tmpdir.join("dummy-file"))
    with open(filpath, 'w') as fil:
        fil.write('foobar')

    spec = skein.ApplicationSpec(
        name="test_master_driver",
        master=skein.Master(
            script='ls\nenv',
            env={'FOO': 'BAR'},
            files={'myfile': filpath}
        )
    )
    with run_application(client, spec=spec, connect=False) as app_id:
        assert wait_for_completion(client, app_id) == 'SUCCEEDED'

    logs = get_logs(client, app_id)
    assert 'FOO=BAR' in logs
    assert 'myfile' in logs


@pytest.mark.parametrize('kind, master_cmd, service_cmd', [
    ('service_succeeds', 'sleep infinity', 'exit 0'),
    ('service_fails', 'sleep infinity', 'exit 1'),
    ('driver_succeeds', 'exit 0', 'sleep infinity'),
    ('driver_fails', 'exit 1', 'sleep infinity')
])
def test_master_driver_shutdown_sequence(kind, master_cmd, service_cmd,
                                         client, tmpdir):
    spec = skein.ApplicationSpec(
        name="test_master_driver_shutdown_sequence_%s" % kind,
        master=skein.Master(script=master_cmd),
        services={
            'service': skein.Service(
                resources=skein.Resources(memory=32, vcores=1),
                script=service_cmd
            )
        }
    )

    state = 'SUCCEEDED' if kind.endswith('succeeds') else 'FAILED'

    if kind == 'service_succeeds':
        with run_application(client, spec=spec) as app:
            wait_for_containers(app, 1, states=['SUCCEEDED'])
            assert len(app.get_containers()) == 0
            # App hangs around until driver completes
            app.shutdown()
            assert wait_for_completion(client, app.id) == state
    else:
        with run_application(client, spec=spec, connect=False) as app_id:
            # service_fails results in immediate failure
            # driver_succeeds results in immediate success
            # driver_fails results in immediate failure
            assert wait_for_completion(client, app_id) == state


test_retries_script_template = """
if [[ $CONTAINER_ID =~ container_[0-9]+_[0-9]+_{succeed_on}_[0-9]+ ]]; then
  echo "Succeeding on attempt {succeed_on}"
  exit 0
else
  echo "Failing on other attempts"
  exit 1
fi
"""


def test_retries_succeeds(client):
    hdfs = pytest.importorskip('pyarrow.hdfs')

    spec = skein.ApplicationSpec(
        name="test_application_retries_succeeds",
        max_attempts=2,
        master=skein.Master(
            script=test_retries_script_template.format(succeed_on='02')
        )
    )
    with run_application(client, spec=spec, connect=False) as app_id:
        assert wait_for_completion(client, app_id) == 'SUCCEEDED'
    logs = get_logs(client, app_id)
    assert 'Failing on other attempts' in logs
    assert 'Application attempt 1 out of 2 failed, will retry' in logs
    assert 'Succeeding on attempt 02' in logs

    fs = hdfs.connect()
    assert not fs.exists("/user/testuser/.skein/%s" % app_id)


def test_retries_fails(client):
    hdfs = pytest.importorskip('pyarrow.hdfs')

    # Global maximum is 2, checks that appmaster uses 2 instead of 10
    max_attempts = 10

    spec = skein.ApplicationSpec(
        name="test_application_retries_fails",
        max_attempts=max_attempts,
        master=skein.Master(
            script=test_retries_script_template.format(succeed_on='03')
        )
    )
    with run_application(client, spec=spec, connect=False) as app_id:
        assert wait_for_completion(client, app_id) == 'FAILED'
    logs = get_logs(client, app_id)
    assert logs.count('Failing on other attempts') == 2
    assert 'Application attempt 1 out of 2 failed' in logs

    fs = hdfs.connect()
    assert not fs.exists("/user/testuser/.skein/%s" % app_id)


@pytest.mark.parametrize('allow_failures', [False, True])
def test_allow_failures_max_restarts(client, allow_failures):
    name = "test_max_restarts_allow_failures_%s" % str(allow_failures).lower()
    spec = skein.ApplicationSpec(
        name=name,
        master=skein.Master(
            script="sleep infinity"
        ),
        services={
            'myservice': skein.Service(
                instances=1,
                max_restarts=2,
                allow_failures=allow_failures,
                resources=skein.Resources(memory=32, vcores=1),
                script="exit 1"
            )
        }
    )
    with run_application(client, spec=spec) as app:
        if allow_failures:
            # Service failed 3 times, doesn't keep trying to run more
            wait_for_containers(app, 3, states=['FAILED'])
            # Check still running fine after 3 failures
            time.sleep(0.5)
            app.get_specification()
            # Shutdown manually
            app.shutdown()
            assert wait_for_completion(client, app.id) == 'SUCCEEDED'
        else:
            # Service failed 3 times and then terminates
            assert wait_for_completion(client, app.id) == 'FAILED'


@pytest.mark.parametrize('with_restarts', [True, False])
def test_fail_on_container_failure(client, with_restarts):
    script = ('if [[ "$SKEIN_CONTAINER_ID" != "test_0" ]]; then\n'
              '  exit 1\n'
              'else\n'
              '  sleep infinity\n'
              'fi')

    spec = skein.ApplicationSpec(
        name="test_fail_on_container_failure",
        services={
            'test': skein.Service(
                instances=2,
                max_restarts=2 if with_restarts else 0,
                resources=skein.Resources(memory=32, vcores=1),
                script=script
            )
        }
    )
    with run_application(client, spec=spec) as app:
        wait_for_completion(client, app.id) == "FAILED"

    logs = get_logs(client, app.id)
    assert "test_0" in logs
    assert "test_1" in logs
    assert ("test_2" in logs) == with_restarts
    assert ("test_3" in logs) == with_restarts
    assert "test_4" not in logs


def test_move_application(client):
    spec = skein.ApplicationSpec(
        name="test_move_application",
        queue="default",
        master=skein.Master(script="sleep infinity")
    )

    def assert_good_message(msg):
        # Ensure message doesn't contain traceback
        assert 'org.apache.hadoop' not in str(msg)

    with run_application(client, spec=spec) as app:
        assert client.application_report(app.id).queue == "default"

        # Successful move
        client.move_application(app.id, "apples")
        assert client.application_report(app.id).queue == "apples"

        # Not a leaf queue
        with pytest.raises(ValueError) as exc:
            client.move_application(app.id, "fruit")
        assert 'Leaf' in str(exc.value)
        assert_good_message(exc.value)

        # Queue doesn't exist
        with pytest.raises(ValueError) as exc:
            client.move_application(app.id, "missing")
        assert "doesn't exist" in str(exc.value)
        assert_good_message(exc.value)

        app.shutdown()

    # App already shutdown
    with pytest.raises(ValueError) as exc:
        client.move_application(app.id, "default")
    assert "cannot be moved" in str(exc.value)
    assert_good_message(exc.value)

    # App doesn't exist
    missing_appid = 'application_1526134340424_0012'
    with pytest.raises(ValueError) as exc:
        client.move_application(missing_appid, "default")
    # This error message is different in Hadoop 3
    assert "absent" in str(exc.value) or "doesn't exist" in str(exc.value)
    assert_good_message(exc.value)

    # Invalid application id
    with pytest.raises(ValueError) as exc:
        client.move_application("oh no", "default")
    assert "Invalid" in str(exc.value)


def test_hadoop3_resource(client):
    spec = skein.ApplicationSpec(
        name="test_hadoop3_resources",
        master=skein.Master(
            resources=skein.Resources(
                memory='32 MiB',
                vcores=1,
                gpus=1
            ),
            script="sleep infinity"
        )
    )
    with pytest.raises(ValueError) as exc:
        client.submit(spec)
    if HADOOP3:
        assert "Resource 'yarn.io/gpu'" in str(exc.value)
    else:
        assert "Custom resources not supported"



import multiprocessing
from datetime import timedelta, datetime, timezone

from skein.utils import (humanize_timedelta, format_table,
                         format_comma_separated_list, lock_file,
                         datetime_from_millis, datetime_to_millis)


def test_humanize_timedelta():
    assert humanize_timedelta(timedelta(0, 6)) == '6s'
    assert humanize_timedelta(timedelta(0, 601)) == '10m'
    assert humanize_timedelta(timedelta(0, 6001)) == '1h 40m'


def test_datetime_conversion():
    # Timezone naive
    now = datetime.now()
    now2 = datetime_from_millis(datetime_to_millis(now))
    now3 = datetime_from_millis(datetime_to_millis(now2))
    # Rounded to nearest millisecond
    assert abs(now2 - now).total_seconds() < 0.001
    assert now2 == now3

    # Timezone aware
    now = datetime.now(timezone.utc)
    now2 = datetime_from_millis(datetime_to_millis(now)).replace(tzinfo=timezone.utc)
    now3 = datetime_from_millis(datetime_to_millis(now2)).replace(tzinfo=timezone.utc)
    # Rounded to nearest millisecond
    assert abs(now2 - now).total_seconds() < 0.001
    assert now2 == now3


def test_format_table():
    res = format_table(['fruit', 'color'],
                       [('apple', 'red'),
                        ('banana', 'yellow'),
                        ('tomato', 'red'),
                        ('pear', 'green')])
    sol = ('FRUIT     COLOR\n'
           'apple     red\n'
           'banana    yellow\n'
           'tomato    red\n'
           'pear      green')
    assert res == sol
    assert format_table(['fruit', 'color'], []) == 'FRUIT    COLOR'


def test_format_comma_separated_list():
    assert format_comma_separated_list([]) == ''
    assert format_comma_separated_list([None]) == 'None'
    assert format_comma_separated_list([None, False]) == 'None or False'
    assert format_comma_separated_list([None, False, True]) == 'None, False, or True'


def test_lock_file_single_process(tmpdir):
    path = str(tmpdir.join('lock'))
    acquired = False
    with lock_file(path):
        acquired = True
    assert acquired


def test_lock_file_same_object_used_for_same_path(tmpdir):
    path = str(tmpdir.join('lock'))
    lock1 = lock_file(path)
    lock2 = lock_file(path)
    assert lock1 is lock2


class Locker(multiprocessing.Process):
    def __init__(self, path=None):
        super(Locker, self).__init__()
        self.path = path
        self.daemon = True
        self.parent_conn, self.child_conn = multiprocessing.Pipe()

    def shutdown(self):
        if not self.parent_conn.closed:
            self.parent_conn.send('shutdown')
            self.parent_conn.close()
        self.join()

    def run(self):
        lock = lock_file(self.path)
        while True:
            msg = self.child_conn.recv()
            if msg == 'acquire':
                lock.acquire()
                self.child_conn.send('acquired')
            elif msg == 'release':
                lock.release()
                self.child_conn.send('released')
            elif msg == 'shutdown':
                break
        self.child_conn.close()


def test_lock_file_multiple_processes(tmpdir):
    path = str(tmpdir.join('lock'))

    lock = lock_file(path)

    locker = Locker(path=path)
    locker.start()
    locker.parent_conn.send('acquire')
    assert 'acquired' == locker.parent_conn.recv()
    locker.parent_conn.send('release')
    assert 'released' == locker.parent_conn.recv()

    lock.acquire()
    locker.parent_conn.send('acquire')
    assert not locker.parent_conn.poll(0.1)
    lock.release()
    assert 'acquired' == locker.parent_conn.recv()
    locker.parent_conn.send('release')
    assert 'released' == locker.parent_conn.recv()
    locker.shutdown()
    locker.join()



import pytest

import skein
from skein.exceptions import ConnectionError
from skein.test.conftest import run_application, wait_for_containers

requests = pytest.importorskip('requests')


simplehttp = skein.Service(resources=skein.Resources(memory=32, vcores=1),
                           script='/usr/bin/python -m SimpleHTTPServer 8888')
master = skein.Master(resources=skein.Resources(memory=256, vcores=1),
                      script="sleep infinity")
spec = skein.ApplicationSpec(name="test_webui",
                             queue="default",
                             master=master,
                             services={'simplehttp': simplehttp})

LOGIN = '?user.name=testuser'
PAGES = [('route-1', 'http://worker.example.com:8888', 'link 1'),
         ('route-2', 'http://worker.example.com:8888/tmp/', 'link 2'),
         ('route-3', 'http://worker.example.com:8888/container_tokens', 'link 3'),
         ('route-4', 'http://worker.example.com:8888', None)]


def get_page(address, **kwargs):
    # Set timeout on all requests so tests don't hang on error
    kwargs.setdefault('timeout', 5)
    return requests.get(address, **kwargs)


@pytest.fixture
def ui_pages(ui_test_app):
    pages = tuple(ui_test_app.ui.add_page(*args) for args in PAGES)
    try:
        yield pages
    finally:
        for page in ui_test_app.ui.get_pages():
            ui_test_app.ui.remove_page(page)


@pytest.fixture(scope="module")
def ui_test_app(client, has_kerberos_enabled):
    if has_kerberos_enabled:
        pytest.skip("Testing only implemented for simple authentication")
    with run_application(client, spec=spec) as app:
        # Wait for a single container
        wait_for_containers(app, 1, states=['RUNNING'])
        try:
            yield app
        finally:
            try:
                app.shutdown()
            except ConnectionError:
                client.kill_application(app.id)


@pytest.mark.parametrize('target, suffix', [
    ('http://example.com:8888', '/'),
    ('http://example.com:8888/', '/'),
    ('http://example.com:8888/foo', ''),
    ('http://example.com:8888/foo/', '/')
])
def test_webui_proxied_pages(target, suffix):
    base = 'http://master.example.com:8088/proxy/application_1526497750451_0001/'
    p = skein.ui.ProxiedPage('test', target, None, base, '/pages')
    assert p.address == base + 'pages/test' + suffix
    assert repr(p) == "ProxiedPage<route='test'>"


def test_webui_basics(ui_test_app):
    assert ui_test_app.ui.proxy_prefix == '/pages'

    assert len(ui_test_app.ui.addresses) == 1

    address = 'http://master.example.com:8088/proxy/%s/' % ui_test_app.id
    assert ui_test_app.ui.address == address

    assert repr(ui_test_app.ui) == "WebUI<address=%r>" % address


def test_webui_authentication(ui_test_app):
    # Fails without authentication
    resp = get_page(ui_test_app.ui.address)
    assert resp.status_code == 401

    # With authentication
    resp = get_page(ui_test_app.ui.address + LOGIN)
    assert resp.ok

    # Cookies set, no longer need authentication
    cookies = resp.cookies
    resp = get_page(ui_test_app.ui.address, cookies=cookies)
    assert resp.ok


def test_webui_home_and_overview(ui_test_app):
    # / and /overview are the same
    for suffix in ['', 'overview']:
        resp = get_page(ui_test_app.ui.address + suffix + LOGIN)
        assert resp.ok
        # Application Metrics
        assert "Memory:</b> 288.0 MiB" in resp.text
        assert "Cores:</b> 2" in resp.text
        assert "Progress:</b> N/A" in resp.text
        # Logs
        assert "testuser/application.master.log" in resp.text  # master
        assert "testuser/application.driver.log" in resp.text  # driver
        assert '/testuser/simplehttp.log' in resp.text  # service
        # Services
        assert 'simplehttp_0' in resp.text  # list of containers


def test_webui_kv(ui_test_app):
    ui_test_app.kv['foo'] = b'bar'
    ui_test_app.kv['bad'] = b'\255\255\255'  # non-unicode

    resp = get_page(ui_test_app.ui.address + 'kv' + LOGIN)
    assert resp.ok
    assert 'foo' in resp.text
    assert 'bar' in resp.text
    assert 'bad' in resp.text
    assert '&lt;binary value&gt;' in resp.text


def test_webui_resources_reachable(ui_test_app):
    resp = get_page(ui_test_app.ui.address + 'favicon.ico' + LOGIN)
    assert resp.ok


def test_webui_404(ui_test_app):
    resp = get_page(ui_test_app.ui.address + 'fake' + LOGIN)
    assert resp.status_code == 404


def test_webui_pages(ui_test_app, ui_pages):
    for page, (route, target, link_name) in zip(ui_pages, PAGES):
        assert page.route == route
        assert page.target == target
        assert page.link_name == link_name

    res = ui_test_app.ui.get_pages()
    sol = {p.route: p for p in ui_pages}
    assert res == sol

    ui_test_app.ui.remove_page(ui_pages[0].route)

    res = ui_test_app.ui.get_pages()
    sol = {p.route: p for p in ui_pages[1:]}
    assert res == sol


def test_webui_pages_errors(ui_test_app, ui_pages):
    route, target, link_name = PAGES[0]

    # Route already used
    with pytest.raises(ValueError):
        ui_test_app.ui.add_page(route, target)

    # Link name already used
    with pytest.raises(ValueError):
        ui_test_app.ui.add_page('new-link', target, link_name=link_name)

    # Invalid url
    with pytest.raises(ValueError):
        ui_test_app.ui.add_page('new-link', 'invalid url')

    # Page doesn't exist
    with pytest.raises(ValueError):
        ui_test_app.ui.remove_page('new-link')


def test_webui_pages_access(ui_test_app, ui_pages):
    root_page, dir_page, file_page, no_link_page = ui_pages

    # Get cookies
    resp = get_page(ui_test_app.ui.address + LOGIN)
    assert resp.ok
    cookies = resp.cookies

    resp = get_page(root_page.address, cookies=cookies)
    assert resp.ok

    resp = get_page(dir_page.address, cookies=cookies)
    assert resp.ok

    resp = get_page(file_page.address, cookies=cookies)
    assert resp.ok

    resp = get_page(root_page.address + "/tmp/", cookies=cookies)
    assert resp.ok

    resp = get_page(root_page.address + "/this-doesnt-exist", cookies=cookies)
    assert resp.status_code == 404

    # 404 after page is removed
    ui_test_app.ui.remove_page(dir_page.route)

    resp = get_page(dir_page.address, cookies=cookies)
    assert resp.status_code == 404


def test_webui_pages_dropdown(ui_test_app, ui_pages):
    root_page, dir_page, file_page, no_link_page = ui_pages

    resp = get_page(ui_test_app.ui.address + LOGIN)
    assert resp.ok

    assert 'href="./pages/%s/"' % root_page.route in resp.text
    assert root_page.link_name in resp.text
    assert 'href="./pages/%s/"' % dir_page.route in resp.text
    assert dir_page.link_name in resp.text
    assert 'href="./pages/%s"' % file_page.route in resp.text
    assert file_page.link_name in resp.text

    assert 'href-"./pages/%s"' % no_link_page.route not in resp.text

    for page in [root_page, dir_page, file_page]:
        ui_test_app.ui.remove_page(page.route)

    resp = get_page(ui_test_app.ui.address + LOGIN)
    assert resp.ok

    assert "No service pages available" in resp.text


def test_webui_no_acls(ui_test_app):
    # No ACLs mean anyone has access
    resp = get_page(ui_test_app.ui.address + "?user.name=baduser")
    assert resp.ok


@pytest.mark.parametrize('ui_users, checks', [
    (['*'], [('testuser', True), ('testuser2', True)]),
    ([], [('testuser', True), ('testuser2', False)]),
    (['testuser', 'testuser2'], [('testuser', True),
                                 ('testuser2', True),
                                 ('testuser3', False)])
])
def test_webui_acls(client, has_kerberos_enabled, ui_users, checks):
    if has_kerberos_enabled:
        pytest.skip("Testing only implemented for simple authentication")

    spec = skein.ApplicationSpec(name="test_webui_acls",
                                 queue="default",
                                 master=master,
                                 acls=skein.ACLs(enable=True, ui_users=ui_users))

    with run_application(client, spec=spec) as app:
        # Base url of web ui
        base = 'http://master.example.com:8088/proxy/%s' % app.id

        # Check proper subset of users allowed
        for user, ok in checks:
            resp = get_page(base + "?user.name=%s" % user)
            assert resp.ok == ok

        app.shutdown()



import copy
import operator
import sys
import threading
import time
import weakref
from collections import OrderedDict
from collections.abc import MutableMapping

import pytest

from skein import kv
from skein.exceptions import ConnectionError
from skein.test.conftest import run_application


@pytest.fixture(scope="module")
def kv_test_app_persistent(client):
    with run_application(client) as app:
        # ensure one container id exists, but already completed
        app.scale('sleeper', 2)
        app.kill_container('sleeper_0')
        try:
            yield app
        finally:
            app.shutdown()


@pytest.fixture
def kv_test_app(kv_test_app_persistent):
    try:
        yield kv_test_app_persistent
    finally:
        kv_test_app_persistent.kv.clear()
        kv_test_app_persistent.scale('sleeper', 1)


kv_test_data = {'bar': b'a',
                'barf': b'b',
                'bars': b'c',
                'bart': b'd',
                'foo': b'e',
                'food': b'f',
                'foodie': b'g'}


def test_clean_tab_completion_kv_namespace():
    public = [n for n in dir(kv) if not n.startswith('_')]
    assert set(public) == set(kv.__all__)


@pytest.mark.parametrize('field', ['value', 'owner'])
def test_field_accessor_types(field):
    cls = getattr(kv, field)
    rec = cls('foo')

    assert repr(rec) == "%s('foo')" % field

    assert not kv.is_condition(rec)

    rec2 = copy.deepcopy(rec)
    assert rec2.key == 'foo'
    assert isinstance(rec2, cls)

    with pytest.raises(TypeError):
        cls(1)


def test_comparison_type():
    # test using `comparison` directly
    v = kv.comparison('foo', 'value', '==', b'bar')
    assert v == (kv.value('foo') == b'bar')

    o = kv.comparison('foo', 'owner', '==', 'bar_1')
    assert o == (kv.owner('foo') == 'bar_1')

    with pytest.raises(TypeError):
        kv.comparison(1, 'value', '==', b'bar')

    with pytest.raises(ValueError):
        kv.comparison('foo', 'unknown', '==', b'bar')

    with pytest.raises(ValueError):
        kv.comparison('foo', 'value', 'unknown', b'bar')

    with pytest.raises(TypeError):
        kv.comparison('foo', 'owner', '<=', None)


@pytest.mark.parametrize('field, rhs1, rhs2, supports_none',
                         [('value', b'rhs1', b'rhs2', False),
                          ('owner', 'c_1', 'c_2', True)])
@pytest.mark.parametrize('op, symbol', [(operator.eq, '=='), (operator.ne, '!='),
                                        (operator.lt, '<'), (operator.le, '<='),
                                        (operator.gt, '>'), (operator.ge, '>=')])
def test_comparison_builder_types(field, rhs1, rhs2, supports_none, op, symbol):
    cls = getattr(kv, field)
    rec = cls('foo')

    x1 = op(rec, rhs1)
    assert kv.is_condition(x1)
    assert isinstance(x1, kv.comparison)
    assert repr(x1) == "%s('foo') %s %r" % (field, symbol, rhs1)

    x2 = op(rec, rhs2)
    assert kv.is_condition(x2)
    assert isinstance(x2, kv.comparison)
    assert repr(x2) == "%s('foo') %s %r" % (field, symbol, rhs2)

    assert x1 == copy.deepcopy(x1)
    assert x1 != x2

    if supports_none and symbol in ('==', '!='):
        op(rec, None)
    else:
        with pytest.raises(TypeError):
            op(rec, None)

    with pytest.raises(TypeError):
        op(rec, 123)

    for attr, val in [('key', 'foo'), ('field', field),
                      ('operator', symbol), ('rhs', rhs1)]:
        assert getattr(x1, attr) == val

        with pytest.raises(AttributeError):
            setattr(x1, attr, val)  # immutable


@pytest.mark.parametrize('cls', [kv.exists,
                                 kv.missing])
def test_exists_and_missing_types(cls):
    x = cls('foo')
    assert repr(x) == ("%s('foo')" % cls.__name__)
    assert x == copy.deepcopy(x)
    assert x != cls('bar')

    assert kv.is_condition(x)
    assert kv.is_operation(x)

    with pytest.raises(AttributeError):
        x.foobar = 1

    with pytest.raises(TypeError):
        cls(1)


@pytest.mark.parametrize('cls', [kv.count, kv.list_keys])
def test_count_and_list_keys_types(cls):
    x1 = cls()
    assert repr(x1) == ("%s(start=None, end=None)" % cls.__name__)
    x2 = cls(prefix='foo')
    assert repr(x2) == ("%s(prefix='foo')" % cls.__name__)
    x3 = cls(start='foo')
    assert repr(x3) == ("%s(start='foo', end=None)" % cls.__name__)
    for x in [x1, x2, x3]:
        assert x == copy.deepcopy(x)
    assert x1 != x2

    assert kv.is_operation(x1)

    with pytest.raises(AttributeError):
        x1.foobar = 1

    with pytest.raises(ValueError):
        cls(prefix='foo', start='bar')

    for k in ['prefix', 'start', 'end']:
        with pytest.raises(TypeError):
            cls(**{k: 1})


def test_discard_type():
    x = kv.discard('foo')
    assert repr(x) == "discard('foo')"
    assert x == copy.deepcopy(x)
    assert x != kv.discard('bar')

    assert kv.is_operation(x)

    with pytest.raises(AttributeError):
        x.foobar = 1

    with pytest.raises(TypeError):
        kv.discard(1)


@pytest.mark.parametrize('cls', [kv.get, kv.pop])
def test_get_and_pop_types(cls):
    name = cls.__name__
    x1 = cls('key')
    assert repr(x1) == ("%s('key', default=None, return_owner=False)" % name)
    x2 = cls('key', default=b'bar', return_owner=True)
    assert repr(x2) == ("%s('key', default=%s, return_owner=True)" % (name, repr(b'bar')))
    for x in [x1, x2]:
        assert x == copy.deepcopy(x)
    assert x1 != x2

    assert kv.is_operation(x1)

    with pytest.raises(AttributeError):
        x1.foobar = 1

    with pytest.raises(TypeError):
        cls('foo', default=1)

    with pytest.raises(TypeError):
        cls('foo', return_owner=None)

    with pytest.raises(TypeError):
        cls(1)


@pytest.mark.parametrize('cls,kw', [(kv.get_prefix, 'return_owner'),
                                    (kv.pop_prefix, 'return_owner'),
                                    (kv.discard_prefix, 'return_keys')])
def test_prefix_types(cls, kw):
    name = cls.__name__
    x1 = cls('key')
    assert repr(x1) == ("%s('key', %s=False)" % (name, kw))
    x2 = cls('key', **{kw: True})
    assert repr(x2) == ("%s('key', %s=True)" % (name, kw))
    for x in [x1, x2]:
        assert x == copy.deepcopy(x)
    assert x1 != x2

    assert kv.is_operation(x1)

    with pytest.raises(AttributeError):
        x1.foobar = 1

    with pytest.raises(TypeError):
        cls(1)

    with pytest.raises(TypeError):
        cls('foo', **{kw: None})


@pytest.mark.parametrize('cls,kw', [(kv.get_range, 'return_owner'),
                                    (kv.pop_range, 'return_owner'),
                                    (kv.discard_range, 'return_keys')])
def test_range_types(cls, kw):
    name = cls.__name__
    x1 = cls()
    assert repr(x1) == ("%s(start=None, end=None, %s=False)" % (name, kw))
    x2 = cls(start='key', end='bar', **{kw: True})
    assert repr(x2) == ("%s(start='key', end='bar', %s=True)" % (name, kw))
    for x in [x1, x2]:
        assert x == copy.deepcopy(x)
    assert x1 != x2

    assert kv.is_operation(x1)

    with pytest.raises(AttributeError):
        x1.foobar = 1

    for k in ['start', 'end', kw]:
        with pytest.raises(TypeError):
            cls(**{k: 1})


@pytest.mark.parametrize('cls,has_return_owner',
                         [(kv.put, False), (kv.swap, True)])
def test_put_and_swap_types(cls, has_return_owner):
    name = cls.__name__
    if has_return_owner:
        extra = ', return_owner=False'
    else:
        extra = ''
    x1 = cls('foo', value=b'bar')
    assert repr(x1) == ("%s('foo', value=%s, owner=no_change%s)"
                        % (name, repr(b'bar'), extra))
    x2 = cls('foo', owner='container_1')
    assert repr(x2) == ("%s('foo', value=no_change, owner='container_1'%s)"
                        % (name, extra))
    x3 = cls('foo', owner=None)
    assert repr(x3) == ("%s('foo', value=no_change, owner=None%s)"
                        % (name, extra))
    for x in [x1, x2, x3]:
        assert x == copy.deepcopy(x)
    assert x1 != x2

    assert kv.is_operation(x1)

    with pytest.raises(AttributeError):
        x1.foobar = 1

    with pytest.raises(ValueError):
        cls('foo', owner='invalid')

    with pytest.raises(TypeError):
        cls(1, value=b'bar')

    with pytest.raises(TypeError):
        cls('foo', owner=1)

    with pytest.raises(ValueError):
        cls('foo')


class MapLike(object):
    def __init__(self, data):
        self.data = data

    def keys(self):
        return self.data.keys()

    def __getitem__(self, key):
        return self.data[key]


def test_key_value_mutablemapping(kv_test_app):
    assert isinstance(kv_test_app.kv, MutableMapping)

    assert kv_test_app.kv is kv_test_app.kv

    # Iterators work when empty
    assert len(kv_test_app.kv) == 0
    assert list(kv_test_app.kv) == []
    assert dict(kv_test_app.kv) == {}

    # __setitem__
    kv_test_app.kv['k1'] = b'v1'
    kv_test_app.kv['k2'] = b'v2'

    with pytest.raises(TypeError):
        kv_test_app.kv['k3'] = u'v3'

    # __contains__
    assert 'k1' in kv_test_app.kv
    assert 'missing' not in kv_test_app.kv

    # __len__
    assert len(kv_test_app.kv) == 2

    # __iter__ (note this is sorted)
    assert list(kv_test_app.kv) == ['k1', 'k2']

    # __getitem__
    assert kv_test_app.kv['k1'] == b'v1'

    with pytest.raises(KeyError):
        kv_test_app.kv['missing']

    with pytest.raises(TypeError):
        kv_test_app.kv[1]

    # __delitem__
    kv_test_app.kv['key'] = b'temp'
    assert 'key' in kv_test_app.kv
    del kv_test_app.kv['key']
    assert 'key' not in kv_test_app.kv

    with pytest.raises(KeyError):
        del kv_test_app.kv['missing']

    with pytest.raises(TypeError):
        del kv_test_app.kv[1]

    # setdefault
    assert kv_test_app.kv.setdefault('new_key', b'val1') == b'val1'
    assert kv_test_app.kv.setdefault('new_key', b'val2') == b'val1'

    # clear
    assert len(kv_test_app.kv) > 0
    kv_test_app.kv.clear()
    assert len(kv_test_app.kv) == 0

    # update
    kv_test_app.kv.update({'a': b'0', 'b': b'1'})            # mapping
    kv_test_app.kv.update([('c', b'2'), ('d', b'3')])        # iterable
    kv_test_app.kv.update(MapLike({'e': b'4', 'f': b'5'}))   # class with keys
    kv_test_app.kv.update({'g': b'bad', 'h': b'7'}, g=b'6')  # kwarg overrides
    kv_test_app.kv.update(i=b'8', j=b'9')                    # kwargs only
    sol = {k: str(i).encode() for i, k in enumerate('abcdefghij')}
    assert kv_test_app.kv.get_range() == sol

    with pytest.raises(TypeError):
        kv_test_app.kv.update({'a': 1}, {'b': 2})


def test_key_value_count(kv_test_app):
    kv_test_app.kv.update(kv_test_data)
    assert kv_test_app.kv.count() == 7
    assert kv_test_app.kv.count(prefix='bar') == 4
    assert kv_test_app.kv.count(start='bars') == 5
    assert kv_test_app.kv.count(end='bars') == 2
    assert kv_test_app.kv.count(start='bars', end='food') == 3


def test_key_value_list_keys(kv_test_app):
    kv_test_app.kv.update(kv_test_data)
    assert kv_test_app.kv.list_keys() == list(sorted(kv_test_data))
    assert (kv_test_app.kv.list_keys(prefix='bar') ==
            ['bar', 'barf', 'bars', 'bart'])
    assert (kv_test_app.kv.list_keys(start='bars') ==
            ['bars', 'bart', 'foo', 'food', 'foodie'])
    assert kv_test_app.kv.list_keys(end='bars') == ['bar', 'barf']
    assert (kv_test_app.kv.list_keys(start='bars', end='food') ==
            ['bars', 'bart', 'foo'])


def test_key_value_exists_and_missing(kv_test_app):
    kv_test_app.kv.update(kv_test_data)
    assert 'bar' in kv_test_app.kv
    assert 'missing' not in kv_test_app.kv
    assert kv_test_app.kv.exists('bar')
    assert not kv_test_app.kv.missing('bar')
    assert not kv_test_app.kv.exists('missing')
    assert kv_test_app.kv.missing('missing')


def test_key_value_discard(kv_test_app):
    kv_test_app.kv.update(kv_test_data)
    assert kv_test_app.kv.discard('bar')
    assert 'bar' not in kv_test_app.kv
    assert not kv_test_app.kv.discard('missing')


def test_key_value_discard_prefix(kv_test_app):
    bars = sorted(k for k in kv_test_data if k.startswith('bar'))

    kv_test_app.kv.update(kv_test_data)
    assert kv_test_app.kv.discard_prefix('bar') == len(bars)
    assert kv_test_app.kv.discard_prefix('bar') == 0
    assert kv_test_app.kv.discard_prefix('missing') == 0

    kv_test_app.kv.update(kv_test_data)
    assert kv_test_app.kv.discard_prefix('bar', return_keys=True) == bars
    assert kv_test_app.kv.discard_prefix('bar', return_keys=True) == []
    assert kv_test_app.kv.discard_prefix('missing', return_keys=True) == []


def test_key_value_discard_range(kv_test_app):
    # start
    kv_test_app.kv.update(kv_test_data)
    sol = sorted(k for k in kv_test_data if 'bar' <= k)
    assert kv_test_app.kv.discard_range(start='bar') == len(sol)
    assert kv_test_app.kv.discard_range(start='bar') == 0
    kv_test_app.kv.update(kv_test_data)
    assert kv_test_app.kv.discard_range(start='bar', return_keys=True) == sol
    assert kv_test_app.kv.discard_range(start='bar', return_keys=True) == []

    # end
    kv_test_app.kv.update(kv_test_data)
    sol = sorted(k for k in kv_test_data if k < 'food')
    assert kv_test_app.kv.discard_range(end='food') == len(sol)
    assert kv_test_app.kv.discard_range(end='food') == 0
    kv_test_app.kv.update(kv_test_data)
    assert kv_test_app.kv.discard_range(end='food', return_keys=True) == sol
    assert kv_test_app.kv.discard_range(end='food', return_keys=True) == []

    # both
    kv_test_app.kv.update(kv_test_data)
    sol = sorted(k for k in kv_test_data if 'bar' <= k < 'food')
    assert kv_test_app.kv.discard_range(start='bar', end='food') == len(sol)
    assert kv_test_app.kv.discard_range(start='bar', end='food') == 0
    kv_test_app.kv.update(kv_test_data)
    assert kv_test_app.kv.discard_range(start='bar', end='food',
                                        return_keys=True) == sol
    assert kv_test_app.kv.discard_range(start='bar', end='food',
                                        return_keys=True) == []

    # neither
    kv_test_app.kv.update(kv_test_data)
    sol = sorted(kv_test_data)
    assert kv_test_app.kv.discard_range(return_keys=True) == sol
    assert kv_test_app.kv.discard_range() == 0

    # empty
    kv_test_app.kv.update(kv_test_data)
    assert kv_test_app.kv.discard_range(start='zzzz') == 0
    assert kv_test_app.kv.discard_range(start='z', end='a') == 0


def test_key_value_get(kv_test_app):
    kv_test_app.kv.update(kv_test_data)

    assert kv_test_app.kv.get('bar') == b'a'
    assert kv_test_app.kv.get('bar', return_owner=True) == (b'a', None)

    assert kv_test_app.kv.get('missing') is None
    assert kv_test_app.kv.get('missing', b'default') == b'default'
    assert kv_test_app.kv.get('missing', return_owner=True) == (None, None)
    assert (kv_test_app.kv.get('missing', b'default', return_owner=True) ==
            (b'default', None))


def test_key_value_get_prefix(kv_test_app):
    empty = OrderedDict()
    kv_test_app.kv.update(kv_test_data)

    sol = OrderedDict(sorted((k, v) for k, v in kv_test_data.items()
                             if k.startswith('bar')))
    assert kv_test_app.kv.get_prefix('bar') == sol
    assert kv_test_app.kv.get_prefix('missing') == empty

    sol = OrderedDict(sorted((k, (v, None)) for k, v in kv_test_data.items()
                             if k.startswith('bar')))
    assert kv_test_app.kv.get_prefix('bar', return_owner=True) == sol
    assert kv_test_app.kv.get_prefix('missing', return_owner=True) == empty


def test_key_value_get_range(kv_test_app):
    kv_test_app.kv.update(kv_test_data)

    # start
    sol = OrderedDict(sorted((k, v) for k, v in kv_test_data.items()
                             if 'bar' <= k))
    assert kv_test_app.kv.get_range(start='bar') == sol
    sol = OrderedDict(sorted((k, (v, None)) for k, v in kv_test_data.items()
                             if 'bar' <= k))
    assert kv_test_app.kv.get_range(start='bar', return_owner=True) == sol

    # end
    sol = OrderedDict(sorted((k, v) for k, v in kv_test_data.items()
                             if k < 'food'))
    assert kv_test_app.kv.get_range(end='food') == sol
    sol = OrderedDict(sorted((k, (v, None)) for k, v in kv_test_data.items()
                             if k < 'food'))
    assert kv_test_app.kv.get_range(end='food', return_owner=True) == sol

    # both
    sol = OrderedDict(sorted((k, v) for k, v in kv_test_data.items()
                             if 'bar' <= k < 'food'))
    assert kv_test_app.kv.get_range(start='bar', end='food') == sol
    sol = OrderedDict(sorted((k, (v, None)) for k, v in kv_test_data.items()
                             if 'bar' <= k < 'food'))
    assert kv_test_app.kv.get_range(start='bar', end='food',
                                    return_owner=True) == sol

    # neither
    sol = OrderedDict(sorted(kv_test_data.items()))
    assert kv_test_app.kv.get_range() == sol

    # empty
    assert kv_test_app.kv.get_range(start='zzzz') == OrderedDict()
    assert kv_test_app.kv.get_range(start='z', end='a') == OrderedDict()


def test_key_value_pop(kv_test_app):
    kv_test_app.kv.update(kv_test_data)

    assert kv_test_app.kv.pop('bar') == b'a'
    assert 'bar' not in kv_test_app.kv

    assert kv_test_app.kv.pop('barf', return_owner=True) == (b'b', None)
    assert 'barf' not in kv_test_app.kv

    assert kv_test_app.kv.pop('missing') is None
    assert kv_test_app.kv.pop('missing', b'default') == b'default'
    assert kv_test_app.kv.pop('missing', return_owner=True) == (None, None)
    assert (kv_test_app.kv.pop('missing', b'default', return_owner=True) ==
            (b'default', None))


def test_key_value_pop_prefix(kv_test_app):
    empty = OrderedDict()

    kv_test_app.kv.update(kv_test_data)
    sol = OrderedDict(sorted((k, v) for k, v in kv_test_data.items()
                             if k.startswith('bar')))
    assert kv_test_app.kv.pop_prefix('bar') == sol
    assert kv_test_app.kv.pop_prefix('bar') == empty
    assert kv_test_app.kv.pop_prefix('missing') == empty

    kv_test_app.kv.update(kv_test_data)
    sol = OrderedDict(sorted((k, (v, None)) for k, v in kv_test_data.items()
                             if k.startswith('bar')))
    assert kv_test_app.kv.pop_prefix('bar', return_owner=True) == sol
    assert kv_test_app.kv.pop_prefix('bar', return_owner=True) == empty
    assert kv_test_app.kv.pop_prefix('missing', return_owner=True) == empty


def test_key_value_pop_range(kv_test_app):
    empty = OrderedDict()

    # start
    kv_test_app.kv.update(kv_test_data)
    sol = OrderedDict(sorted((k, v) for k, v in kv_test_data.items()
                             if 'bar' <= k))
    assert kv_test_app.kv.pop_range(start='bar') == sol
    assert kv_test_app.kv.pop_range(start='bar') == empty
    kv_test_app.kv.update(kv_test_data)
    sol = OrderedDict(sorted((k, (v, None)) for k, v in kv_test_data.items()
                             if 'bar' <= k))
    assert kv_test_app.kv.pop_range(start='bar', return_owner=True) == sol
    assert kv_test_app.kv.pop_range(start='bar', return_owner=True) == empty

    # end
    kv_test_app.kv.update(kv_test_data)
    sol = OrderedDict(sorted((k, v) for k, v in kv_test_data.items()
                             if k < 'food'))
    assert kv_test_app.kv.pop_range(end='food') == sol
    assert kv_test_app.kv.pop_range(end='food') == empty
    kv_test_app.kv.update(kv_test_data)
    sol = OrderedDict(sorted((k, (v, None)) for k, v in kv_test_data.items()
                             if k < 'food'))
    assert kv_test_app.kv.pop_range(end='food', return_owner=True) == sol
    assert kv_test_app.kv.pop_range(end='food', return_owner=True) == empty

    # both
    kv_test_app.kv.update(kv_test_data)
    sol = OrderedDict(sorted((k, v) for k, v in kv_test_data.items()
                             if 'bar' <= k < 'food'))
    assert kv_test_app.kv.pop_range(start='bar', end='food') == sol
    assert kv_test_app.kv.pop_range(start='bar', end='food') == empty
    kv_test_app.kv.update(kv_test_data)
    sol = OrderedDict(sorted((k, (v, None)) for k, v in kv_test_data.items()
                             if 'bar' <= k < 'food'))
    assert kv_test_app.kv.pop_range(start='bar', end='food',
                                    return_owner=True) == sol
    assert kv_test_app.kv.pop_range(start='bar', end='food',
                                    return_owner=True) == empty

    # neither
    kv_test_app.kv.update(kv_test_data)
    sol = OrderedDict(sorted(kv_test_data.items()))
    assert kv_test_app.kv.pop_range() == sol
    assert kv_test_app.kv.pop_range() == empty

    # empty
    kv_test_app.kv.update(kv_test_data)
    assert kv_test_app.kv.pop_range(start='zzzz') == empty
    assert kv_test_app.kv.pop_range(start='z', end='a') == empty


def test_key_value_put(kv_test_app):
    cid = kv_test_app.get_containers()[0].id

    # New, value only
    kv_test_app.kv.put('foo', value=b'a')
    assert kv_test_app.kv.get('foo', return_owner=True) == (b'a', None)

    # New, owner and value
    kv_test_app.kv.put('bar', value=b'a', owner=cid)
    assert kv_test_app.kv.get('bar', return_owner=True) == (b'a', cid)

    # Overwrite existing key
    kv_test_app.kv.put('foo', value=b'b')
    assert kv_test_app.kv.get('foo', return_owner=True) == (b'b', None)

    # Set owner only, value remains unchanged
    kv_test_app.kv.put('foo', owner=cid)
    assert kv_test_app.kv.get('foo', return_owner=True) == (b'b', cid)

    # Set value only, owner remains unchanged
    kv_test_app.kv.put('foo', value=b'c')
    assert kv_test_app.kv.get('foo', return_owner=True) == (b'c', cid)

    # Clear owner, value remains unchanged
    kv_test_app.kv.put('foo', owner=None)
    assert kv_test_app.kv.get('foo', return_owner=True) == (b'c', None)

    # Set both value and owner
    kv_test_app.kv.put('foo', value=b'd', owner=cid)
    assert kv_test_app.kv.get('foo', return_owner=True) == (b'd', cid)

    # Set value, clear owner
    kv_test_app.kv.put('foo', value=b'e', owner=None)
    assert kv_test_app.kv.get('foo', return_owner=True) == (b'e', None)

    # Can't create new key without setting value as well
    with pytest.raises(ValueError) as exc:
        kv_test_app.kv.put('missing', owner=cid)
    assert str(exc.value) == "ignore_value=True & key isn't already set"

    # Owner service doesn't exist
    with pytest.raises(ValueError) as exc:
        kv_test_app.kv.put('bar', owner='unknown_0')
    assert str(exc.value) == "Unknown service 'unknown'"

    # Owner instance doesn't exist
    with pytest.raises(ValueError) as exc:
        kv_test_app.kv.put('bar', owner='sleeper_100')
    assert str(exc.value) == "Service 'sleeper' has no container instance 100"

    # Owner instance already completed
    with pytest.raises(ValueError) as exc:
        kv_test_app.kv.put('bar', owner='sleeper_0')
    assert str(exc.value) == "Container 'sleeper_0' has already completed"


def test_key_value_swap(kv_test_app):
    cid = kv_test_app.get_containers()[0].id

    # New, return value only
    assert kv_test_app.kv.swap('bar', value=b'a') is None

    # Existing, return value only
    assert kv_test_app.kv.swap('bar', value=b'b') == b'a'

    # Return owner for remaining tests
    ro = {'return_owner': True}

    # New, value only
    assert kv_test_app.kv.swap('foo', value=b'a', **ro) == (None, None)
    assert kv_test_app.kv.get('foo', **ro) == (b'a', None)

    # Overwrite existing key
    assert kv_test_app.kv.swap('foo', value=b'b', **ro) == (b'a', None)
    assert kv_test_app.kv.get('foo', **ro) == (b'b', None)

    # Set owner only, value remains unchanged
    assert kv_test_app.kv.swap('foo', owner=cid, **ro) == (b'b', None)
    assert kv_test_app.kv.get('foo', **ro) == (b'b', cid)

    # Set value only, owner remains unchanged
    assert kv_test_app.kv.swap('foo', value=b'c', **ro) == (b'b', cid)
    assert kv_test_app.kv.get('foo', **ro) == (b'c', cid)

    # Clear owner, value remains unchanged
    assert kv_test_app.kv.swap('foo', owner=None, **ro) == (b'c', cid)
    assert kv_test_app.kv.get('foo', **ro) == (b'c', None)

    # Set both value and owner
    assert (kv_test_app.kv.swap('foo', value=b'd', owner=cid, **ro) ==
            (b'c', None))
    assert kv_test_app.kv.get('foo', **ro) == (b'd', cid)

    # Set value, clear owner
    assert (kv_test_app.kv.swap('foo', value=b'e', owner=None, **ro) ==
            (b'd', cid))
    assert kv_test_app.kv.get('foo', return_owner=True) == (b'e', None)

    # Can't create new key without setting value as well
    with pytest.raises(ValueError) as exc:
        kv_test_app.kv.swap('missing', owner=cid)
    assert str(exc.value) == "ignore_value=True & key isn't already set"

    # Owner service doesn't exist
    with pytest.raises(ValueError) as exc:
        kv_test_app.kv.swap('bar', owner='unknown_0')
    assert str(exc.value) == "Unknown service 'unknown'"

    # Owner instance doesn't exist
    with pytest.raises(ValueError) as exc:
        kv_test_app.kv.swap('bar', owner='sleeper_100')
    assert str(exc.value) == "Service 'sleeper' has no container instance 100"

    # Owner instance already completed
    with pytest.raises(ValueError) as exc:
        kv_test_app.kv.swap('bar', owner='sleeper_0')
    assert str(exc.value) == "Container 'sleeper_0' has already completed"


def test_key_value_ownership(kv_test_app):
    kv_test_app.scale('sleeper', 2)
    c1, c2 = (c.id for c in kv_test_app.get_containers())

    # Create some owned keys
    kv_test_app.kv.put('c1_1', value=b'a', owner=c1)
    kv_test_app.kv.put('c1_2', value=b'b', owner=c1)
    kv_test_app.kv.put('c1_3', value=b'c', owner=c1)
    # Change the owners for two of them
    kv_test_app.kv.put('c1_2', owner=c2)
    kv_test_app.kv.put('c1_3', owner=None)

    state = kv_test_app.kv.get_range(return_owner=True)
    assert state['c1_1'] == (b'a', c1)
    assert state['c1_2'] == (b'b', c2)
    assert state['c1_3'] == (b'c', None)

    kv_test_app.kill_container(c1)

    # c1_1 should be deleted, the others should remain
    assert not kv_test_app.kv.exists('c1_1')
    assert kv_test_app.kv.exists('c1_2')
    assert kv_test_app.kv.exists('c1_3')


def test_key_value_keys_fully_removed_from_previous_owners(kv_test_app):
    # Tests changing an owner, removing an owner, and deletion all fully remove
    # a key from being owned by a container.
    kv_test_app.scale('sleeper', 2)
    c1, c2 = (c.id for c in kv_test_app.get_containers())

    # Create some owned keys
    kv_test_app.kv.put('c1_1', value=b'a', owner=c1)
    kv_test_app.kv.put('c1_2', value=b'b', owner=c1)
    kv_test_app.kv.put('c1_3', value=b'c', owner=c1)

    # Change the owners for two of them
    kv_test_app.kv.put('c1_2', owner=c2)
    kv_test_app.kv.put('c1_3', owner=None)

    # Delete all keys
    kv_test_app.kv.clear()

    # Recreate the keys with no owners
    kv_test_app.kv.update({'c1_1': b'a', 'c1_2': b'b', 'c1_3': b'c'})

    # Kill the original owning container
    kv_test_app.kill_container(c1)

    # Check that the keys still exist
    for k in ['c1_1', 'c1_2', 'c1_3']:
        assert k in kv_test_app.kv


def test_transaction_conditions(kv_test_app):
    cid = kv_test_app.get_containers()[0].id
    service, instance = cid.split('_')
    instance = int(instance)
    cid_larger = '%s_%d' % (service, instance + 1)
    cid_smaller = '%s_%d' % (service, instance - 1)

    val = b'bcd'
    larger = b'bce'
    smaller = b'bcc'

    kv_test_app.kv.put('owner', val, owner=cid)
    kv_test_app.kv.put('no_owner', val)

    def run(*conds):
        return kv_test_app.kv.transaction(conds).succeeded

    # -- value --
    # exists
    assert run(kv.exists('owner'))
    assert not run(kv.exists('missing'))
    assert run(kv.missing('missing'))
    assert not run(kv.missing('owner'))

    # ==
    assert run(kv.value('owner') == val)
    assert not run(kv.value('owner') == larger)
    assert not run(kv.value('missing') == val)
    # !=
    assert run(kv.value('owner') != larger)
    assert not run(kv.value('owner') != val)
    assert run(kv.value('missing') != val)
    # <
    assert run(kv.value('owner') < larger)
    assert not run(kv.value('owner') < val)
    assert not run(kv.value('owner') < smaller)
    assert not run(kv.value('missing') < val)
    # <=
    assert run(kv.value('owner') <= larger)
    assert run(kv.value('owner') <= val)
    assert not run(kv.value('owner') <= smaller)
    assert not run(kv.value('missing') <= val)
    # >
    assert not run(kv.value('owner') > larger)
    assert not run(kv.value('owner') > val)
    assert run(kv.value('owner') > smaller)
    assert not run(kv.value('missing') > val)
    # >=
    assert not run(kv.value('owner') >= larger)
    assert run(kv.value('owner') >= val)
    assert run(kv.value('owner') >= smaller)
    assert not run(kv.value('missing') >= val)

    # -- owner --
    # Owner presence/absence
    assert run(kv.owner('owner') != None)  # noqa
    assert not run(kv.owner('owner') == None)  # noqa
    assert run(kv.owner('no_owner') == None)  # noqa
    assert not run(kv.owner('no_owner') != None)  # noqa
    # ==
    assert run(kv.owner('owner') == cid)
    assert not run(kv.owner('owner') == cid_larger)
    assert not run(kv.owner('no_owner') == cid)
    # !=
    assert run(kv.owner('owner') != cid_larger)
    assert not run(kv.owner('owner') != cid)
    assert run(kv.owner('no_owner') != cid)
    # <
    assert run(kv.owner('owner') < cid_larger)
    assert not run(kv.owner('owner') < cid)
    assert not run(kv.owner('owner') < cid_smaller)
    assert not run(kv.owner('no_owner') < cid)
    # <=
    assert run(kv.owner('owner') <= cid_larger)
    assert run(kv.owner('owner') <= cid)
    assert not run(kv.owner('owner') <= cid_smaller)
    assert not run(kv.owner('no_owner') <= cid)
    # >
    assert not run(kv.owner('owner') > cid_larger)
    assert not run(kv.owner('owner') > cid)
    assert run(kv.owner('owner') > cid_smaller)
    assert not run(kv.owner('no_owner') > cid)
    # >=
    assert not run(kv.owner('owner') >= cid_larger)
    assert run(kv.owner('owner') >= cid)
    assert run(kv.owner('owner') >= cid_smaller)
    assert not run(kv.owner('no_owner') >= cid)

    # No conditions
    assert run()

    # Multiple conditions
    true1 = kv.exists('owner')
    true2 = kv.exists('no_owner')
    false = kv.exists('missing')
    assert run(true1, true2)
    assert not run(false, true1)
    assert not run(true1, false, true2)


@pytest.mark.parametrize('op, msg', [
    (kv.put('missing', owner='sleeper_0'), "ignore_value=True & key isn't already set"),
    (kv.put('a', owner='unknown_0'), "Unknown service 'unknown'"),
    (kv.put('a', owner='sleeper_99'), "Service 'sleeper' has no container instance 99"),
    (kv.put('a', owner='sleeper_0'), "Container 'sleeper_0' has already completed")])
def test_transaction_put_key_errors(kv_test_app, op, msg):
    cid = kv_test_app.get_containers()[0].id
    kv_test_app.kv.put('a', b'a', owner=cid)

    for succeeded in [True, False]:
        if succeeded:
            kwargs = {'conditions': [kv.exists('a')],
                      'on_success': [kv.put('b', b'b'), op],
                      'on_failure': [kv.put('c', b'c')]}
        else:
            kwargs = {'conditions': [kv.missing('a')],
                      'on_success': [kv.put('c', b'c')],
                      'on_failure': [kv.put('b', b'b'), op]}

        with pytest.raises(ValueError) as exc:
            kv_test_app.kv.transaction(**kwargs)

        assert str(exc.value) == msg
        assert not kv_test_app.kv.exists('b')
        assert not kv_test_app.kv.exists('c')


def test_transaction(kv_test_app):
    cid = kv_test_app.get_containers()[0].id

    kv_test_app.kv.put('a', b'a', owner=cid)
    kv_test_app.kv.put('aa', b'aa')
    kv_test_app.kv.put('aaa', b'aaa')
    kv_test_app.kv.put('b', b'b')

    true1 = kv.exists('a')
    true2 = kv.exists('b')
    false = kv.missing('a')

    # Test building results in success and failure
    for succeeded in [True, False]:
        if succeeded:
            cond, do, dont = [true1, true2], 'on_success', 'on_failure'
        else:
            cond, do, dont = [true1, false], 'on_failure', 'on_success'

        kwargs = {'conditions': cond,
                  do: [kv.get_prefix('a'), kv.get('missing')],
                  dont: [kv.put('dont_put_me', b'bad')]}

        res = kv_test_app.kv.transaction(**kwargs)

        assert res.succeeded == succeeded
        assert len(res.results) == 2
        assert res.results[0] == kv_test_app.kv.get_prefix('a')
        assert res.results[1] == kv_test_app.kv.get('missing')
        assert not kv_test_app.kv.exists('dont_put_me')

        # Test empty result branch
        kwargs = {'conditions': cond, dont: [kv.put('dont_put_me', b'bad')]}
        res = kv_test_app.kv.transaction(**kwargs)
        assert res == (succeeded, [])
        assert not kv_test_app.kv.exists('dont_put_me')

    # Test one operation of each type
    a_prefixes = kv_test_app.kv.get_prefix('a')

    res = kv_test_app.kv.transaction(
        conditions=[true1, true2],
        on_success=[kv.get_prefix('a'),
                    kv.discard_prefix('a', return_keys=True),
                    kv.put('a', b'later')])

    assert res.succeeded
    assert len(res.results) == 3
    assert res.results[0] == a_prefixes
    assert res.results[1] == list(a_prefixes)
    assert res.results[2] is None

    # Test no operations
    assert kv_test_app.kv.transaction() == (True, [])

    # invalid argument types
    with pytest.raises(TypeError):
        kv_test_app.kv.transaction(conditions=[kv.get('foo')])

    with pytest.raises(TypeError):
        kv_test_app.kv.transaction(on_success=[kv.value('foo') == b'bar'])

    with pytest.raises(TypeError):
        kv_test_app.kv.transaction(on_failure=[kv.value('foo') == b'bar'])


def test_event_filter_type():
    x = kv.EventFilter()
    assert x.start is None
    assert x.end is None
    assert x.event_type == kv.EventType.ALL

    assert copy.copy(x) == x
    assert hash(x) == hash(x)

    # smoketest
    repr(x)

    x2 = kv.EventFilter(prefix='foo')
    assert x2.start == 'foo'
    assert x2.end == 'fop'
    assert kv.EventFilter(prefix='foo') == x2
    assert kv.EventFilter(prefix='bar') != x2

    x3 = kv.EventFilter(key='foo')
    assert x3.start == 'foo'
    assert x3.end == 'foo\x00'
    assert kv.EventFilter(key='foo') == x3
    assert kv.EventFilter(key='bar') != x3

    assert kv.EventFilter(event_type='delete').event_type == 'DELETE'

    for k in ['key', 'prefix', 'start', 'end', 'event_type']:
        with pytest.raises(TypeError):
            kv.EventFilter(**{k: 1})

    with pytest.raises(ValueError):
        kv.EventFilter(key='foo', prefix='bar')

    with pytest.raises(ValueError):
        kv.EventFilter(event_type='foo')


def test_event_queue(kv_test_app):
    q = kv_test_app.kv.events(prefix='foo')
    assert len(q.filters) == 1
    e_foo = next(iter(q.filters))

    # smoketest
    repr(q)

    # Already subscribed
    assert q.subscribe(prefix='foo') == e_foo
    assert len(q.filters) == 1

    # subscribe/unsubscribe with keywords
    e_bar = q.subscribe(key='bar')
    assert len(q.filters) == 2
    assert e_bar in q.filters
    q.unsubscribe(key='bar')
    assert len(q.filters) == 1
    assert e_bar not in q.filters

    # subscribe/unsubscribe with explicit filter
    e_baz = kv.EventFilter(key='baz')
    assert q.subscribe(e_baz) == e_baz
    assert e_baz in q.filters
    assert q.unsubscribe(e_baz) == e_baz
    assert e_baz not in q.filters

    with pytest.raises(ValueError):
        q.unsubscribe(key='missing')

    with pytest.raises(ValueError):
        q.unsubscribe(event_filter=e_foo, prefix='bar')

    with pytest.raises(TypeError):
        q.unsubscribe(event_filter='incorrect type')

    # Context manager
    with q:
        q.subscribe(prefix='more')
        assert len(q.filters) == 2
    assert len(q.filters) == 0

    # Test queue methods
    q.put(1)
    assert q.get() == 1

    # Queue is iterable
    q.put(1)
    q.put(2)
    iq = iter(q)
    next(iq) == 1
    next(iq) == 2

    # Test queue methods with errors
    q.put(ValueError("foobar"))
    for _ in range(2):
        with pytest.raises(ValueError) as exc:
            q.get()
        assert str(exc.value) == 'foobar'


def test_event_queue_cleanup(kv_test_app):
    q = kv_test_app.kv.events(prefix='foo')

    # Put a few events in the queue
    kv_test_app.kv['foo'] = b'1'
    del kv_test_app.kv['foo']

    # No extra references
    assert sys.getrefcount(q) == 2

    # del causes queue to be collected
    assert len(kv_test_app.kv._filter_to_queues) == 1
    ref = weakref.ref(q)
    del q
    assert ref() is None
    assert len(kv_test_app.kv._filter_to_queues) == 0


def get_n(q, n):
    return [q.get(timeout=2) for _ in range(n)]


def put_event(k, v, o, filter):
    return kv.Event(k, kv.ValueOwnerPair(v, o), kv.EventType.PUT, filter)


def del_event(k, filter):
    return kv.Event(k, None, kv.EventType.DELETE, filter)


def test_event_ranges(kv_test_app):
    before_d = kv_test_app.kv.events(end='d')
    after_c = kv_test_app.kv.events(start='c')
    b_to_e = kv_test_app.kv.events(start='b', end='e')
    everything = kv_test_app.kv.events()

    kv_test_app.kv.put('a', b'1')
    kv_test_app.kv.put('b', b'2')
    kv_test_app.kv.put('c', b'3')
    kv_test_app.kv.put('d', b'4')
    kv_test_app.kv.put('e', b'5')
    kv_test_app.kv.put('f', b'6')
    kv_test_app.kv.discard_range(start='b', end='d')
    kv_test_app.kv.discard_range(end='d')
    kv_test_app.kv.discard_range(start='e')
    kv_test_app.kv.discard('d')
    assert len(kv_test_app.kv) == 0

    # Open start
    filt = next(iter(before_d.filters))
    sol = [put_event('a', b'1', None, filt),
           put_event('b', b'2', None, filt),
           put_event('c', b'3', None, filt),
           del_event('b', filt),
           del_event('c', filt),
           del_event('a', filt)]
    res = get_n(before_d, len(sol))
    assert res == sol

    # Open end
    filt = next(iter(after_c.filters))
    sol = [put_event('c', b'3', None, filt),
           put_event('d', b'4', None, filt),
           put_event('e', b'5', None, filt),
           put_event('f', b'6', None, filt),
           del_event('c', filt),
           del_event('e', filt),
           del_event('f', filt),
           del_event('d', filt)]
    res = get_n(after_c, len(sol))
    assert res == sol

    # Closed start and end
    filt = next(iter(b_to_e.filters))
    sol = [put_event('b', b'2', None, filt),
           put_event('c', b'3', None, filt),
           put_event('d', b'4', None, filt),
           del_event('b', filt),
           del_event('c', filt),
           del_event('d', filt)]
    res = get_n(b_to_e, len(sol))
    assert res == sol

    # Open start and end
    filt = next(iter(everything.filters))
    sol = [put_event('a', b'1', None, filt),
           put_event('b', b'2', None, filt),
           put_event('c', b'3', None, filt),
           put_event('d', b'4', None, filt),
           put_event('e', b'5', None, filt),
           put_event('f', b'6', None, filt),
           del_event('b', filt),
           del_event('c', filt),
           del_event('a', filt),
           del_event('e', filt),
           del_event('f', filt),
           del_event('d', filt)]
    res = get_n(everything, len(sol))
    assert res == sol


def test_event_types(kv_test_app):
    put_b_to_e = kv_test_app.kv.events(start='b',
                                       end='e',
                                       event_type='put')
    del_b_to_e = kv_test_app.kv.events(start='b',
                                       end='e',
                                       event_type='delete')

    kv_test_app.kv.put('a', b'1')
    kv_test_app.kv.put('b', b'2')
    kv_test_app.kv.put('c', b'3')
    kv_test_app.kv.put('d', b'4')
    kv_test_app.kv.put('e', b'5')
    kv_test_app.kv.discard_range(start='b', end='d')
    kv_test_app.kv.discard_range(end='d')
    kv_test_app.kv.discard_range(start='e')
    kv_test_app.kv.clear()

    # put
    filt = next(iter(put_b_to_e.filters))
    sol = [put_event('b', b'2', None, filt),
           put_event('c', b'3', None, filt),
           put_event('d', b'4', None, filt)]
    res = get_n(put_b_to_e, len(sol))
    assert res == sol

    # del
    filt = next(iter(del_b_to_e.filters))
    sol = [del_event('b', filt),
           del_event('c', filt),
           del_event('d', filt)]
    res = get_n(del_b_to_e, len(sol))
    assert res == sol


def test_event_multiple_subscriptions(kv_test_app):
    q1 = kv_test_app.kv.event_queue()
    before_b = q1.subscribe(end='b')
    after_c = q1.subscribe(start='c')

    q2 = kv_test_app.kv.event_queue()
    assert q2.subscribe(end='b') == before_b
    after_c_del = q2.subscribe(start='c', event_type='delete')

    kv_test_app.kv.put('a', b'1')
    kv_test_app.kv.put('b', b'2')
    kv_test_app.kv.put('c', b'3')
    kv_test_app.kv.put('d', b'4')
    kv_test_app.kv.discard_range(start='b', end='d')
    kv_test_app.kv.discard_range(end='d')
    kv_test_app.kv.discard_range(start='e')
    kv_test_app.kv.clear()

    sol = [put_event('a', b'1', None, before_b),
           put_event('c', b'3', None, after_c),
           put_event('d', b'4', None, after_c),
           del_event('c', after_c),
           del_event('a', before_b),
           del_event('d', after_c)]
    res = get_n(q1, len(sol))
    assert res == sol

    sol = [put_event('a', b'1', None, before_b),
           del_event('c', after_c_del),
           del_event('a', before_b),
           del_event('d', after_c_del)]
    res = get_n(q2, len(sol))
    assert res == sol


def test_events_and_ownership(kv_test_app):
    kv_test_app.scale('sleeper', 2)
    c1, c2 = (c.id for c in kv_test_app.get_containers())

    q1 = kv_test_app.kv.events(prefix='a')
    q2 = kv_test_app.kv.events(prefix='a', event_type='put')

    # Create some owned keys
    kv_test_app.kv.put('a1', value=b'1', owner=c1)
    kv_test_app.kv.put('a2', value=b'2', owner=c1)
    kv_test_app.kv.put('a3', value=b'3', owner=c2)
    kv_test_app.kv.put('a4', value=b'4', owner=c2)

    # Change owner of some
    kv_test_app.kv.put('a3', owner=c1)
    kv_test_app.kv.put('a4', owner=None)

    # Delete an owned key
    del kv_test_app.kv['a1']

    # Kill c1
    kv_test_app.kill_container(c1)

    # Delete all remaining keys
    kv_test_app.kv.clear()

    filt = next(iter(q1.filters))
    sols = [[put_event('a1', b'1', c1, filt),
             put_event('a2', b'2', c1, filt),
             put_event('a3', b'3', c2, filt),
             put_event('a4', b'4', c2, filt),
             put_event('a3', b'3', c1, filt),
             put_event('a4', b'4', None, filt)] +
            [del_event('a1', filt),
             del_event(x, filt),
             del_event(y, filt),
             del_event('a4', filt)]
            for x, y in [('a2', 'a3'), ('a3', 'a2')]]
    res = get_n(q1, len(sols[0]))
    assert res in sols

    filt = next(iter(q2.filters))
    sol = [put_event('a1', b'1', c1, filt),
           put_event('a2', b'2', c1, filt),
           put_event('a3', b'3', c2, filt),
           put_event('a4', b'4', c2, filt),
           put_event('a3', b'3', c1, filt),
           put_event('a4', b'4', None, filt)]
    res = get_n(q2, len(sol))
    assert res == sol


def test_key_value_wait(kv_test_app):
    def set_foo():
        time.sleep(0.5)
        kv_test_app.kv['foo'] = b'1'

    for return_owner in [True, False]:
        kv_test_app.kv.clear()

        sol = (b'1', None) if return_owner else b'1'

        setter = threading.Thread(target=set_foo)
        setter.daemon = True
        setter.start()

        val = kv_test_app.kv.wait('foo', return_owner=return_owner)
        assert val == sol

        # Get immediately for set keys
        val = kv_test_app.kv.wait('foo', return_owner=return_owner)
        assert val == sol


def test_events_application_shutdown(client):
    with run_application(client) as app:
        q = app.kv.events(prefix='a')
        app.kv.put('a1', b'1')
        app.kv.put('a2', b'2')
        app.kv.put('a3', b'3')
        app.shutdown()

    filt = next(iter(q.filters))
    assert q.get() == put_event('a1', b'1', None, filt)
    assert q.get() == put_event('a2', b'2', None, filt)
    assert q.get() == put_event('a3', b'3', None, filt)

    # All further requests error with connection error
    for _ in range(2):
        with pytest.raises(ConnectionError):
            q.get()



import errno
import os
import re
import subprocess
import sys
import pkg_resources
from distutils.command.build import build as _build
from distutils.command.clean import clean as _clean
from distutils.dir_util import remove_tree
from glob import glob

from setuptools import setup, Command
from setuptools.command.develop import develop as _develop
from setuptools.command.install import install as _install

import versioneer

VERSION = versioneer.get_version()

ROOT_DIR = os.path.abspath(os.path.dirname(os.path.relpath(__file__)))
JAVA_DIR = os.path.join(ROOT_DIR, 'java')
JAVA_TARGET_DIR = os.path.join(JAVA_DIR, 'target')
JAVA_PROTO_DIR = os.path.join(ROOT_DIR, "java", "src", "main", "proto")
SKEIN_JAVA_DIR = os.path.join(ROOT_DIR, 'skein', 'java')
SKEIN_JAR = os.path.join(SKEIN_JAVA_DIR, 'skein.jar')
SKEIN_PROTO_DIR = os.path.join(ROOT_DIR, 'skein', 'proto')


class build_proto(Command):
    description = "build protobuf artifacts"

    user_options = []

    def initialize_options(self):
        pass

    finalize_options = initialize_options

    def _fix_imports(self, path):
        new = ['from __future__ import absolute_import']
        with open(path) as fil:
            for line in fil:
                if re.match("^import [^ ]*_pb2 as [^ ]*$", line):
                    line = 'from . ' + line
                new.append(line)

        with open(path, 'w') as fil:
            fil.write(''.join(new))

    def run(self):
        from grpc_tools import protoc
        include = pkg_resources.resource_filename('grpc_tools', '_proto')
        for src in glob(os.path.join(JAVA_PROTO_DIR, "*.proto")):
            command = ['grpc_tools.protoc',
                       '--proto_path=%s' % JAVA_PROTO_DIR,
                       '--proto_path=%s' % include,
                       '--python_out=%s' % SKEIN_PROTO_DIR,
                       '--grpc_python_out=%s' % SKEIN_PROTO_DIR,
                       src]
            if protoc.main(command) != 0:
                self.warn('Command: `%s` failed'.format(command))
                sys.exit(1)

        for path in _compiled_protos():
            self._fix_imports(path)


class build_java(Command):
    description = "build java artifacts"

    user_options = []

    def initialize_options(self):
        pass

    finalize_options = initialize_options

    def run(self):
        # Compile the java code and copy the jar to skein/java/
        # This will be picked up as package_data later
        self.mkpath(SKEIN_JAVA_DIR)
        try:
            code = subprocess.call(['mvn', '-f', os.path.join(JAVA_DIR, 'pom.xml'),
                                    '-Dskein.version=%s' % VERSION,
                                    '--batch-mode', 'package'])
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                self.warn("Building Skein requires Maven, which wasn't found "
                          "in your environment. For information on setting "
                          "up a build environment for Skein see "
                          "https://jcrist.github.io/skein/develop.html.")
                sys.exit(1)
            raise
        if code:
            sys.exit(code)

        jar_files = glob(os.path.join(JAVA_TARGET_DIR, 'skein-*.jar'))
        if not jar_files:
            self.warn('Maven compilation produced no jar files')
            sys.exit(1)
        elif len(jar_files) > 1:
            self.warn('Maven produced multiple jar files')
            sys.exit(1)

        jar = jar_files[0]

        self.copy_file(jar, SKEIN_JAR)


def _ensure_java(command):
    if not getattr(command, 'no_java', False) and not os.path.exists(SKEIN_JAR):
        command.run_command('build_java')


def _compiled_protos():
    return glob(os.path.join(SKEIN_PROTO_DIR, '*_pb2*.py'))


def _ensure_proto(command):
    if not _compiled_protos():
        command.run_command('build_proto')


class build(_build):
    def run(self):
        _ensure_java(self)
        _ensure_proto(self)
        _build.run(self)


class install(_install):
    def run(self):
        _ensure_java(self)
        _ensure_proto(self)
        _install.run(self)


class develop(_develop):
    user_options = list(_develop.user_options)
    user_options.append(('no-java', None, "Don't build the java source"))

    def initialize_options(self):
        self.no_java = False
        _develop.initialize_options(self)

    def run(self):
        if not self.uninstall:
            _ensure_java(self)
            _ensure_proto(self)
        _develop.run(self)


class clean(_clean):
    def run(self):
        if self.all:
            for d in [SKEIN_JAVA_DIR, JAVA_TARGET_DIR]:
                if os.path.exists(d):
                    remove_tree(d, dry_run=self.dry_run)
            for fil in _compiled_protos():
                if not self.dry_run:
                    os.unlink(fil)
        _clean.run(self)


is_build_step = bool({'build', 'install', 'develop',
                      'bdist_wheel'}.intersection(sys.argv))
protos_built = bool(_compiled_protos()) and 'clean' not in sys.argv

if 'build_proto' in sys.argv or (is_build_step and not protos_built):
    setup_requires = ['grpcio-tools']
else:
    setup_requires = []


install_requires = ['grpcio>=1.11.0',
                    'protobuf>=3.5.0',
                    'pyyaml',
                    'cryptography']

# Due to quirks in setuptools/distutils dependency ordering, to get the java
# and protobuf sources to build automatically in most cases, we need to check
# for them in multiple locations. This is unfortunate, but seems necessary.
cmdclass = versioneer.get_cmdclass()
cmdclass.update({'build_java': build_java,    # directly build the java source
                 'build_proto': build_proto,  # directly build the proto source
                 'build': build,              # bdist_wheel or pip install .
                 'install': install,          # python setup.py install
                 'develop': develop,          # python setup.py develop
                 'clean': clean})             # extra cleanup


setup(name='skein',
      version=VERSION,
      cmdclass=cmdclass,
      maintainer='Jim Crist',
      maintainer_email='jiminy.crist@gmail.com',
      license='BSD',
      description=('A simple tool and library for deploying applications on '
                   'Apache YARN'),
      long_description=(open('README.rst').read()
                        if os.path.exists('README.rst') else ''),
      url='http://github.com/jcrist/skein/',
      classifiers=["Development Status :: 5 - Production/Stable",
                   "License :: OSI Approved :: BSD License",
                   "Programming Language :: Java",
                   "Programming Language :: Python :: 3.5",
                   "Programming Language :: Python :: 3.6",
                   "Programming Language :: Python :: 3.7",
                   "Topic :: Software Development :: Libraries :: Java Libraries",
                   "Topic :: System :: Systems Administration",
                   "Topic :: System :: Distributed Computing"],
      keywords='YARN HDFS hadoop distributed cluster',
      packages=['skein', 'skein.proto', 'skein.recipes'],
      package_data={'skein': ['java/*.jar']},
      entry_points='''
        [console_scripts]
        skein=skein.cli:main
      ''',
      install_requires=install_requires,
      setup_requires=setup_requires,
      python_requires=">=3.5",
      zip_safe=False)



import random
from collections import namedtuple
from urllib.parse import urlsplit

from . import proto
from .objects import Base
from .utils import cached_property


__all__ = ('ProxiedPage', 'WebUI')


class ProxiedPage(Base):
    """A page proxied by the Skein Web UI.

    Attributes
    ----------
    route : string
        The route used in the address to the proxied page.
    target : string
        The target address of the proxy.
    link_name : string or None
        The name of the linked page in the Web UI. If ``None``, the page is
        still proxied but isn't linked to in the Web UI.
    address : string
        The full proxied address to this page.
    """
    __slots__ = ('route', 'target', 'link_name', '_ui_address', '_proxy_prefix')

    def __init__(self, route, target, link_name, _ui_address, _proxy_prefix):
        self.route = route
        self.target = target
        self.link_name = link_name
        self._ui_address = _ui_address
        self._proxy_prefix = _proxy_prefix

    def __repr__(self):
        return 'ProxiedPage<route=%r>' % self.route

    @property
    def address(self):
        """The full proxied address to this page"""
        path = urlsplit(self.target).path
        suffix = '/' if not path or path.endswith('/') else ''
        return '%s%s/%s%s' % (self._ui_address[:-1], self._proxy_prefix,
                              self.route, suffix)


UIInfo = namedtuple('UIInfo', ['addresses', 'address', 'proxy_prefix'])


class WebUI(object):
    """The Skein WebUI."""
    def __init__(self, client):
        self._client = client

    @cached_property
    def _ui_info(self):
        resp = self._client._call('UiInfo', proto.UIInfoRequest())
        addresses = tuple(resp.ui_address)
        return UIInfo(addresses, random.choice(addresses), resp.proxy_prefix)

    @property
    def proxy_prefix(self):
        """The path between the Web UI address and the proxied pages."""
        return self._ui_info.proxy_prefix

    @property
    def addresses(self):
        """All known addresses of the Web UI.

        In most cases this will be a list of length 1. If YARN is running in
        high availability mode, there may be several available addresses. Use
        the ``address`` attribute to choose one at random."""
        return self._ui_info.addresses

    @property
    def address(self):
        """The address of the Web UI.

        If YARN is running in high availability mode, a single address is
        chosen at random on first access."""
        return self._ui_info.address

    def __repr__(self):
        return "WebUI<address=%r>" % self.address

    def add_page(self, route, target, link_name=None):
        """Add a new proxied page to the Web UI.

        Parameters
        ----------
        route : str
            The route for the proxied page. Must be a valid path *segment* in a
            url (e.g. ``foo`` in ``/foo/bar/baz``). Routes must be unique
            across the application.
        target : str
            The target address to be proxied to this page. Must be a valid url.
        link_name : str, optional
            If provided, will be the link text used in the Web UI. If not
            provided, the page will still be proxied, but no link will be added
            to the Web UI. Link names must be unique across the application.

        Returns
        -------
        ProxiedPage
        """
        req = proto.Proxy(route=route, target=target, link_name=link_name)
        self._client._call('AddProxy', req)
        return ProxiedPage(route, target, link_name, self.address,
                           self.proxy_prefix)

    def remove_page(self, route):
        """Remove a proxied page from the Web UI.

        Parameters
        ----------
        route : str
            The route for the proxied page. Must be a valid path *segment* in a
            url (e.g. ``foo`` in ``/foo/bar/baz``). Routes must be unique
            across the application.
        """
        req = proto.RemoveProxyRequest(route=route)
        self._client._call('RemoveProxy', req)

    def get_pages(self):
        """Get all registered pages.

        Returns
        -------
        pages : dict
            A ``dict`` of ``route`` to ``ProxiedPage`` for all pages.
        """
        resp = self._client._call('GetProxies', proto.GetProxiesRequest())
        return {i.route: ProxiedPage(i.route, i.target,
                                     i.link_name if i.link_name else None,
                                     self.address, self.proxy_prefix)
                for i in resp.proxy}



import errno
import fcntl
import os
import threading
import time
import weakref
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from distutils.version import LooseVersion

try:
    from grpc import __version__ as GRPC_VERSION
except ImportError:
    from grpc._grpcio_metadata import __version__ as GRPC_VERSION


_paths_lock = threading.Lock()
_paths_to_locks = weakref.WeakValueDictionary()


def lock_file(path):
    """File based lock on ``path``.

    Creates a file based lock. When acquired, other processes or threads are
    prevented from acquiring the same lock until it is released.
    """
    with _paths_lock:
        lock = _paths_to_locks.get(path)
        if lock is None:
            _paths_to_locks[path] = lock = _FileLock(path)
    return lock


class _FileLock(object):
    """Internal file based lock object"""
    def __init__(self, path):
        self._path = path
        self._file = None
        self._threadlock = threading.Lock()

    def _acquire_file(self):
        if self._file is None:
            self._file = open(self._path, "wb")
        fcntl.flock(self._file, fcntl.LOCK_EX)

    def _release_file(self):
        fcntl.flock(self._file, fcntl.LOCK_UN)
        self._file.close()
        self._file = None

    def acquire(self):
        self._threadlock.acquire()
        try:
            self._acquire_file()
        except Exception:  # pragma: no cover
            self._threadlock.release()
            raise

    def release(self):
        self._release_file()
        self._threadlock.release()

    def __enter__(self):
        self.acquire()

    def __exit__(self, *args):
        self.release()


def pid_exists(pid):
    try:
        os.kill(pid, 0)
    except OSError as exc:
        if exc.errno == errno.ESRCH:
            return False
    return True


@contextmanager
def grpc_fork_support_disabled():
    """Temporarily disable fork support in gRPC.

    Fork + exec has always been supported, but the recent fork handling code in
    gRPC (>= 1.15) results in extraneous error logs currently. For now we
    explicitly disable fork support for gRPC clients we create.
    """
    if LooseVersion(GRPC_VERSION) < '1.18.0':
        key = 'GRPC_ENABLE_FORK_SUPPORT'
        try:
            os.environ[key] = '0'
            yield
        finally:
            del os.environ[key]
    else:
        yield


def xor(a, b):
    return bool(a) != bool(b)


def ensure_unicode(x):
    if type(x) is not str:
        x = x.decode('utf-8')
    return x


def format_list(x):
    return "\n".join("- %s" % s for s in sorted(x))


def format_comma_separated_list(x, conjunction='or'):
    n = len(x)
    if n == 0:
        return ''
    if n == 1:
        return str(x[0])
    if n == 2:
        left, right = x
        return '%s %s %s' % (left, conjunction, right)
    left, right = ', '.join(map(str, x[:-1])), x[-1]
    return '%s, %s %s' % (left, conjunction, right)


def humanize_timedelta(td):
    """Pretty-print a timedelta in a human readable format."""
    secs = int(td.total_seconds())
    hours, secs = divmod(secs, 60 * 60)
    mins, secs = divmod(secs, 60)
    if hours:
        return '%dh %dm' % (hours, mins)
    if mins:
        return '%dm' % mins
    return '%ds' % secs


_EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)


def datetime_from_millis(x):
    """Convert milliseconds since the epoch to a naive `datetime.datetime`"""
    if x is None or x == 0:
        return None
    return datetime.fromtimestamp(x / 1e3)


def datetime_to_millis(x):
    """Convert a `datetime.datetime` to milliseconds since the epoch"""
    if x is None:
        return None
    if hasattr(x, 'timestamp'):
        # Python >= 3.3
        secs = x.timestamp()
    elif x.tzinfo is None:
        # Timezone naive
        secs = (time.mktime((x.year, x.month, x.day,
                             x.hour, x.minute, x.second,
                             -1, -1, -1)) + x.microsecond / 1e6)
    else:
        # Timezone aware
        secs = (x - _EPOCH).total_seconds()
    return int(secs * 1000)


def runtime(start_time, finish_time):
    if start_time is None:
        return timedelta(0)
    if finish_time is None:
        return datetime.now() - start_time
    return finish_time - start_time


def format_table(columns, rows):
    """Formats an ascii table for given columns and rows.

    Parameters
    ----------
    columns : list
        The column names
    rows : list of tuples
        The rows in the table. Each tuple must be the same length as
        ``columns``.
    """
    rows = [tuple(str(i) for i in r) for r in rows]
    columns = tuple(str(i).upper() for i in columns)
    if rows:
        widths = tuple(max(max(map(len, x)), len(c))
                       for x, c in zip(zip(*rows), columns))
    else:
        widths = tuple(map(len, columns))
    row_template = ('    '.join('%%-%ds' for _ in columns)) % widths
    header = (row_template % tuple(columns)).strip()
    if rows:
        data = '\n'.join((row_template % r).strip() for r in rows)
        return '\n'.join([header, data])
    else:
        return header


class cached_property(object):

    def __init__(self, func):
        self.__doc__ = getattr(func, "__doc__")
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self

        res = obj.__dict__[self.func.__name__] = self.func(obj)
        return res


def implements(f):
    def decorator(g):
        g.__doc__ = f.__doc__
        return g
    return decorator



import os
import time
import subprocess
from contextlib import contextmanager

import pytest

import skein


@contextmanager
def set_skein_config(tmpdir):
    tmpdir = str(tmpdir)
    old = skein.properties.config_dir
    try:
        skein.properties._mapping['config_dir'] = tmpdir
        yield tmpdir
    finally:
        skein.properties._mapping['config_dir'] = old


@pytest.fixture
def skein_config(tmpdir_factory):
    with set_skein_config(tmpdir_factory.mktemp('config')) as config:
        yield config


@pytest.fixture(scope="session")
def security(tmpdir_factory):
    path = str(tmpdir_factory.mktemp('security'))
    return skein.Security.new_credentials().to_directory(path)


@pytest.fixture(scope="session")
def has_kerberos_enabled():
    return HAS_KERBEROS


@pytest.fixture(scope="session")
def hadoop3():
    return HADOOP3


KEYTAB_PATH = "/home/testuser/testuser.keytab"
HAS_KERBEROS = os.environ.get('HADOOP_TESTING_CONFIG', '').lower() == 'kerberos'
HADOOP3 = os.environ.get('HADOOP_TESTING_VERSION', '').lower() == 'cdh6'


def do_kinit():
    subprocess.check_call(["kinit", "-kt", KEYTAB_PATH, "testuser"])


@pytest.fixture(scope="session")
def kinit():
    if HAS_KERBEROS:
        do_kinit()


@pytest.fixture
def not_logged_in():
    if not HAS_KERBEROS:
        pytest.skip("Without kerberos, users are always logged in")
    try:
        subprocess.check_call(["kdestroy"])
        yield
    finally:
        do_kinit()


@pytest.fixture(scope="session")
def client(security, kinit):
    with skein.Client(security=security) as client:
        yield client


sleeper = skein.Service(resources=skein.Resources(memory=32, vcores=1),
                        script='sleep infinity')


sleep_until_killed = skein.ApplicationSpec(name="sleep_until_killed",
                                           queue="default",
                                           tags={'sleeps'},
                                           services={'sleeper': sleeper})


def check_is_shutdown(client, app_id, status=None):
    timeleft = 5
    while timeleft:
        if client.application_report(app_id).state != 'RUNNING':
            break
        time.sleep(0.1)
        timeleft -= 0.1
    else:
        assert False, "Application wasn't properly terminated"

    if status is not None:
        assert client.application_report(app_id).final_status == status


def wait_for_completion(client, app_id, timeout=30):
    while timeout:
        final_status = client.application_report(app_id).final_status
        if final_status != 'UNDEFINED':
            return final_status
        time.sleep(0.1)
        timeout -= 0.1
    else:
        assert False, "Application timed out"


@contextmanager
def ensure_shutdown(client, app_id, status=None):
    try:
        yield
    except Exception:
        client.kill_application(app_id)
        raise
    finally:
        try:
            check_is_shutdown(client, app_id, status=status)
        except AssertionError:
            client.kill_application(app_id)
            raise


@contextmanager
def run_application(client, spec=sleep_until_killed, connect=True):
    if connect:
        app = client.submit_and_connect(spec)
        app_id = app.id
    else:
        app_id = app = client.submit(spec)
    with ensure_shutdown(client, app_id):
        yield app


def wait_for_containers(app, n, **kwargs):
    timeleft = 5
    while timeleft:
        containers = app.get_containers(**kwargs)
        if len(containers) == n:
            break
        time.sleep(0.1)
        timeleft -= 0.1
    else:
        assert False, "timeout"

    return containers


def get_logs(client, app_id, tries=20, user=''):
    for i in range(tries):
        try:
            return client.application_logs(app_id, user=user).dumps()
        except Exception:
            if i == tries - 1:
                raise
        time.sleep(0.20)
